/***************************************************************************
 *
 * Copyright (C) 2001 International Business Machines
 * All rights reserved.
 *
 * This file is part of the GPFS mmfslinux kernel module.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer. 
 *  2. Redistributions in binary form must reproduce the above copyright 
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution. 
 *  3. The name of International Business Machines may not be used to endorse 
 *     or promote products derived from this software without specific prior 
 *     written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR 
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Alternatively, provided that this notice is retained in full, this
 * software may be distributed under the terms of the GNU General
 * Public License ("GPL") as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later version,
 * in which case the provisions of the GPL apply INSTEAD OF those given above.
 *
 ****************************************************************************/
/* @(#)39       1.173.3.2  src/avs/fs/mmfs/ts/kernext/gpl-linux/kx.c, mmfs, avs_rhrz2, rhrz2s003a 8/2/12 19:40:38 */

#include <Shark-gpl.h>

#ifdef MODULE
#include <linux/module.h>
#endif
#include <linux/string.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/dcache.h>
#include <linux/file.h>
#include <linux/stat.h>
#include <linux/personality.h>
#include <linux/fcntl.h>
#ifdef GANESHA
#include <linux/sunrpc/svc.h>
#include <linux/nfsd/nfsfh.h>
#include <linux/sunrpc/cache.h>
#include <linux/nfsd/export.h>
#include <linux/fsnotify.h>
#include <linux/quotaops.h>
#include <linux/highuid.h>
#endif
#include <gpfs_fcntl.h>
#include <gpfs_nfs.h>


#ifndef GPFS_ARCH_X86_64
/* grab stuff for madvise */
#define __KERNEL_SYSCALLS__
#endif
#include <linux/unistd.h>
#include <linux/mman.h>

#include <verdep.h>
#include <cxiSystem.h>
#include <cxi2gpfs.h>
#include <cxiCred.h>

#include <linux2gpfs.h>
#include <Trace.h>
#include <cxiMode.h>
#include <cxiTSFattr.h>
#include <cxiVFSStats.h>

#define MAX_NLINK16 65535

#define _SYS_STAT_H
#include <cxiTypes.h>

#ifdef API_32BIT

/* Equivalent structure for stat64 at 32-bit user level (for API remapping) */
/* gcc doesn't align double words with i386/ia32 like it does for ia64, so
 * we need to adjust for this. */
/* Ideally we would include a header file with this definition (ia32.h) instead
   of defining it ourselves, but this typically breaks things
   (such as with x86_64). */
typedef struct stat64_user32 {
  UInt64 st_dev;
  UInt32 __pad1;
  UInt32 __st_ino;
  UInt32 st_mode;
  UInt32 st_nlink;
  UInt32 st_uid;
  UInt32 st_gid;
  UInt64 st_rdev;
  UInt32 __pad2;
  UInt32 st_size_lower;     /* break up st_size to avoid extra padding */
  UInt32 st_size_upper;     /* w/ little Endian ordering */
  UInt32 st_blksize;
  UInt64 st_blocks;
  UInt32 st_atime;
  UInt32 __unused1;
  UInt32 st_mtime;
  UInt32 __unused2;
  UInt32 st_ctime;
  UInt32 __unused3;
  UInt64 st_ino;
} stat64_user32;
#endif /* API_32BIT */

/* check if inode belongs to GPFS */
#define GPFS_TYPE(IP) (!(IP) ? false : \
                        (!(IP->i_sb) ? false : \
                          (!(IP->i_sb->s_type) ? false : \
                            !strcmp(IP->i_sb->s_type->name, "gpfs"))))

int
kxPoll(void* fdListP, int nFDs, int timeoutMS)
{
  return -ENOSYS;
}

int cleanupFD = 0;

#ifdef API_32BIT
static void
vstat32(cxiVattr_t *vattrp, struct stat64_user32 *statbuf)
{
  ENTER(0);
  memset(statbuf, 0, sizeof(struct stat64_user32));

  statbuf->st_dev         = vattrp->va_dev;
  statbuf->st_ino         = vattrp->va_ino;
  statbuf->st_mode        = vattrp->va_mode;
  statbuf->st_nlink       = vattrp->va_nlink;
  statbuf->st_uid         = vattrp->va_uid;
  statbuf->st_gid         = vattrp->va_gid;
  statbuf->st_rdev        = vattrp->va_rdev;
  statbuf->st_size_upper  = High32(vattrp->va_size);
  statbuf->st_size_lower  = Low32(vattrp->va_size);
  statbuf->st_atime       = vattrp->va_atime.tv_sec;
  statbuf->st_mtime       = vattrp->va_mtime.tv_sec;
  statbuf->st_ctime       = vattrp->va_ctime.tv_sec;
  statbuf->st_blksize     = vattrp->va_blocksize;
  statbuf->st_blocks      = vattrp->va_blocks;
  /* STAT64_HAS_BROKEN_ST_INO is set on i386 ... */
  statbuf->__st_ino       = vattrp->va_ino;
  EXIT(0);
}
#endif /* API_32BIT */

void static
vstat(cxiVattr_t *vattrp, struct stat64 *statbuf)
{
  ENTER(0);
  memset(statbuf, 0, sizeof(struct stat64));

  statbuf->st_dev         = vattrp->va_dev;
  statbuf->st_ino         = vattrp->va_ino;
  statbuf->st_mode        = vattrp->va_mode;
  /* Make sure we do a best effort to convert nlink to fit the field */
  statbuf->st_nlink = vattrp->va_nlink;
  if (sizeof(statbuf->st_nlink) == 2)
    statbuf->st_nlink = MIN(vattrp->va_nlink, MAX_NLINK16);
  statbuf->st_uid         = vattrp->va_uid;
  statbuf->st_gid         = vattrp->va_gid;
  statbuf->st_rdev        = vattrp->va_rdev;
  statbuf->st_size        = vattrp->va_size;
  statbuf->st_atime       = vattrp->va_atime.tv_sec;
  statbuf->st_mtime       = vattrp->va_mtime.tv_sec;
  statbuf->st_ctime       = vattrp->va_ctime.tv_sec;
#ifdef STAT_HAVE_NSEC
  statbuf->st_atime_nsec  = vattrp->va_atime.tv_nsec;
  statbuf->st_mtime_nsec  = vattrp->va_mtime.tv_nsec;
  statbuf->st_ctime_nsec  = vattrp->va_ctime.tv_nsec;
#else
  statbuf->st_atime_nsec  = 0;
  statbuf->st_mtime_nsec  = 0;
  statbuf->st_ctime_nsec  = 0;
#endif
  statbuf->st_blksize     = vattrp->va_blocksize;
  statbuf->st_blocks      = vattrp->va_blocks;

#ifdef STAT64_HAS_BROKEN_ST_INO
  /* Linux has 2 struct stat64 definitions:
      1) /usr/include/asm/stat.h
      2) /usr/include/bits/stat.h
     Of course, they differ
      1) st_dev & st_rdev is 2 bytes in (1) and 8 bytes in (2),
         but the 2 definitions overlap.
      2) st_ino is 8 bytes in (1) and 4 bytes in (2)
         and they are in different places!
    Fortunately, (1) defines an ifdef telling us to assign st_ino
    to a second variable which just happens to exactly match the
    definition in (2). */
  statbuf->__st_ino       = vattrp->va_ino;
#endif
  EXIT(0);
}

#ifdef GPFS_ARCH_PPC64
void static
vstat_ppc64(cxiVattr_t *vattrp, struct stat *statbuf)
{
  ENTER(0);
  memset(statbuf, 0, sizeof(struct stat));
                                                                                
  statbuf->st_dev         = vattrp->va_dev;
  statbuf->st_ino         = vattrp->va_ino;
  statbuf->st_mode        = vattrp->va_mode;
  /* Make sure we do a best effort to convert nlink to fit the field */
  statbuf->st_nlink = vattrp->va_nlink;
  if (sizeof(statbuf->st_nlink) == 2)
    statbuf->st_nlink = MIN(vattrp->va_nlink, MAX_NLINK16);
  statbuf->st_uid         = vattrp->va_uid;
  statbuf->st_gid         = vattrp->va_gid;
  statbuf->st_rdev        = vattrp->va_rdev;
  statbuf->st_size        = vattrp->va_size;
  statbuf->st_atime       = vattrp->va_atime.tv_sec;
  statbuf->st_mtime       = vattrp->va_mtime.tv_sec;
  statbuf->st_ctime       = vattrp->va_ctime.tv_sec;
#ifdef STAT_HAVE_NSEC
  statbuf->st_atime_nsec  = vattrp->va_atime.tv_nsec;
  statbuf->st_mtime_nsec  = vattrp->va_mtime.tv_nsec;
  statbuf->st_ctime_nsec  = vattrp->va_ctime.tv_nsec;
#else
  statbuf->st_atime_nsec  = 0;
  statbuf->st_mtime_nsec  = 0;
  statbuf->st_ctime_nsec  = 0;
#endif
  statbuf->st_blksize     = vattrp->va_blocksize;
  statbuf->st_blocks      = vattrp->va_blocks;
  EXIT(0);
}
#endif /* GPFS_ARCH_PPC64 */

#define LITEMASK_EXACT_ATIME(m)   (0  != ((m) & 0x10))
#define LITEMASK_SET_EXACT(m)     (m) |= 0x77
#define LITEMASK_SET_NORMAL(m)    (m)  = 0x67

/* a call to return exact file status information
 * using the file's pathname;
 */
int
tsstat(const char *pathname, struct stat64 *statP, 
       unsigned int *litemaskP, unsigned int statFlags)
{
  struct gpfsVfsData_t *privVfsP;
  cxiNode_t *cnP;
  int code = 0;
  int rc = 0;
  struct nameidata nd;
  struct inode *iP;
  Boolean relePath = false;
  cxiVattr_t vattr;
  int flags = 0;
  unsigned int litemask = 0;
#ifdef GPFS_ARCH_PPC64
  struct stat statbuf;
#else
  struct stat64 statbuf;
#endif
#ifdef API_32BIT
  stat64_user32 statbuf32;
#endif

  ENTER(0);
  if (pathname == NULL || statP == NULL)
  {
    code = 1;
    rc = EINVAL;
    goto xerror;
  }

  if (statFlags & GPFS_NOFOLLOW)
    rc = USER_LPATH(pathname, &nd);
  else
    rc = USER_PATH(pathname, &nd);

  if (rc)
  {
    rc = -rc;
    code = 2;
    goto xerror;
  }
  relePath = true;

  iP = NDP_TO_IP(&nd);
  DBGASSERT(iP != NULL);

  TRACE2(TRACE_VNODE, 1, TRCID_LINUXOPS_TSSTAT_ENTER,
         "tsstat enter: iP 0x%lX statP 0x%lX\n", iP, statP);

  if (!GPFS_TYPE(iP))
  {
    code = 3;
    rc = EINVAL;
    goto xerror;
  }

  privVfsP = VP_TO_PVP(iP);
  if (privVfsP == NULL)
  {
    code = 4;
    rc = ENOENT;
    goto xerror;
  }
  cnP = VP_TO_CNP(iP);

  rc = cxiCopyIn((char *)litemaskP, (char *)&litemask, sizeof(unsigned int));
  if (rc)
  {
    code = 6;
    goto xerror;
  }

  /* Based on litemask, set either GETATTR_EXACT, GETATTR_LITE, or ZERO */
  if (LITEMASK_EXACT_ATIME(litemask))
  {
    /* Accurate atime required exact stat */
    flags = GETATTR_EXACT;
    LITEMASK_SET_EXACT(litemask);
  }
  else if (0 == litemask)
  {
    /* No attrs must be exact.  This is statlite. */
    flags = GETATTR_LITE;
  }
  else
  {
    /* Normal stat (atime need not be accurate) */
    flags = 0;
    litemask = 0x67;
    LITEMASK_SET_NORMAL(litemask);
  }

  rc = gpfs_ops.gpfsGetattr(privVfsP, cnP, &vattr, flags);

  vattr.va_rdev = cxiDev32ToDev(vattr.va_rdev);

  if (rc == 0)
  {
#ifdef API_32BIT
    /* Need to remap? */
    if (cxiIS64U((char *)statP))
    {
      vstat(&vattr, &statbuf);

      rc = cxiCopyOut((char *)&statbuf, (char *)statP, sizeof(struct stat64));
    }
    else
    {
      /* Need to remap for 32-bit API call */
      vstat32(&vattr, &statbuf32);

      rc = cxiCopyOut((char *)&statbuf32, (char *)statP,
                      sizeof(struct stat64_user32));
    }
#else

#ifdef GPFS_ARCH_PPC64
    /* On ppc64, struct stat64 layout is different for a 64-bit user app and
       kernel, and we can't use it for copying data out.  However, struct stat
       has the same layout in kernel and 64-bit userspace, and is identical
       to struct stat64 in 64-bit userspace, so we use it for copyout. */
    vstat_ppc64(&vattr, &statbuf);
#else
    vstat(&vattr, &statbuf);
#endif
    rc = cxiCopyOut((char *)&statbuf, (char *)statP, sizeof(statbuf));

#endif /* API_32BIT */
    if (rc != 0)
    {
      code = 5;
      goto xerror;
    }

    /* return litemask */
    rc = cxiCopyOut((char *)&litemask, (char *)litemaskP, sizeof(unsigned int));
    if (rc)
    {
      code = 7;
      goto xerror;
    }
  }

xerror:
  if (relePath)
    PATH_PUT(&nd);

  TRACE2(TRACE_VNODE, 1, TRCID_LINUXOPS_TSSTAT_EXIT,
         "tsstat exit: code %d rc %d\n", code, rc);
  EXIT(0);
  return -rc;
}

int
tsfstat(int fileDesc, struct stat64 *statP)
{
  struct gpfsVfsData_t *privVfsP;
  cxiNode_t *cnP;
  struct file *fP = NULL;
  struct inode *iP;
  int code = 0;
  int rc;
  Boolean releFile = false;
  cxiVattr_t vattr;
#ifdef GPFS_ARCH_PPC64
  struct stat statbuf;
#else
  struct stat64 statbuf;
#endif
#ifdef API_32BIT
  struct stat64_user32 statbuf32;
#endif /* API_32BIT */

  ENTER(0);
  if (statP == NULL)
  {
    code = 1;
    rc = EINVAL;
    goto xerror;
  }

  fP = fget(fileDesc);
  if (!fP)
  {
    code = 2;
    rc = EBADF;
    goto xerror;
  }
  releFile = true;

  iP = fP->f_dentry->d_inode;
  DBGASSERT(iP != NULL);

  TRACE4(TRACE_VNODE, 1, TRCID_LINUXOPS_TSFSTAT_ENTER,
         "tsfstat enter: fd %d fP 0x%lX iP 0x%lX statP 0x%lX\n",
         fileDesc, fP, iP, statP);

  if (!GPFS_TYPE(iP))
  {
    code = 3;
    rc = EINVAL;
    goto xerror;
  }

  privVfsP = VP_TO_PVP(iP);
  if (privVfsP == NULL)
  {
    code = 4;
    rc = ENOENT;
    goto xerror;
  }
  cnP = VP_TO_CNP(iP);

  rc = gpfs_ops.gpfsGetattr(privVfsP, cnP, &vattr, GETATTR_EXACT);
  vattr.va_rdev = cxiDev32ToDev(vattr.va_rdev);

  if (rc == 0)
  {
#ifdef API_32BIT
    /* Need to remap? */
    if (cxiIS64U((char *)statP))
    {
      vstat(&vattr, &statbuf);

      rc = cxiCopyOut((char *)&statbuf, (char *)statP, sizeof(struct stat64));
    }
    else
    {
      /* Need to remap for 32-bit API call */
      vstat32(&vattr, &statbuf32);

      rc = cxiCopyOut((char *)&statbuf32, (char *)statP,
                      sizeof(struct stat64_user32));
    }
#else

#ifdef GPFS_ARCH_PPC64
    /* On ppc64, struct stat64 layout is different for a 64-bit user app and
       kernel, and we can't use it for copying data out.  However, struct stat
       has the same layout in kernel and 64-bit userspace, and is identical
       to struct stat64 in 64-bit userspace, so we use it for copyout. */
    vstat_ppc64(&vattr, &statbuf);
#else
    vstat(&vattr, &statbuf);
#endif
    rc = cxiCopyOut((char *)&statbuf, (char *)statP, sizeof(statbuf));

#endif /* API_32BIT */
    if (rc != 0)
    {
      code = 5;
      goto xerror;
    }
  }

xerror:
  if (releFile)
    fput(fP);

  TRACE2(TRACE_VNODE, 1, TRCID_LINUXOPS_TSFSTAT_EXIT,
         "tsfstat exit: code %d rc %d\n", code, rc);
  EXIT(0);
  return -rc;
}

int
tsfattr(int fileDesc, int command, void *argP,
        tsfattrReasonCodeInfo *rCodeP)
{
  struct gpfsVfsData_t *privVfsP;
  cxiNode_t *cnP;
  struct MMFSVInfo *vinfoP = NULL;
  ext_cred_t eCred;
  ext_cred_t *eCredP = &eCred;
  struct file *fP = NULL;
  struct inode *iP;
  int reason = 0;
  int code = 0;
  int flags;
  int rc;
  Boolean releFile = false;
  genericStruct_t   *gncp = NULL;
  gpfsFcntlHeader_t *hdrp = NULL;  
    
    
  VFS_STAT_START(tsfattrCall);
  ENTER(0);

#ifdef SMB_LOCKS_TEST
  /* hack to test setSMBOplock via tsfattr call */
  if (command == 999)
  {
    int args[2];
    int oplockGranted;
    struct file *fP = NULL;
    /*??KernelOperation kop;*/

    rc = cxiCopyIn((char *)argP, (char *)args, sizeof(args));
    if (rc)
      return -rc;

    /* Obtain the file pointer from the descriptor */
    fP = fget(fileDesc);
    if (!fP)
      return -EBADF;

    /*??rc = (int)kop.kBegin(privVfsP, CHECK_UNMOUNTED, pudNone, generalOp);*/
    rc = setSMBOplock(fP, fP->private_data, args[0], args[1],
                      (void *)0xdeadbeef, &oplockGranted, false);

    fput(fP);

    if (rc)
      return -rc;
    return oplockGranted;
  }
#endif

  /* if rCodeP was specified then immediately reset reason field */
  if (rCodeP != NULL)
  {
    rc = cxiCopyOut((char*)&reason, (char*)&(rCodeP->reason), sizeof(reason));
    if (rc != 0)
    {
      code = 1;
      goto xerror;
    }
  }

  fP = fget(fileDesc);
  if (!fP)
  {
    code = 2;
    rc = EBADF;
    goto xerror;
  }
  releFile = true;

  iP = fP->f_dentry->d_inode;
  DBGASSERT(iP != NULL);

  flags = cxiOpenFlagsXlate(fP->f_flags);

  TRACE9(TRACE_VNODE, 1, TRCID_LINUXOPS_TSFATTR_ENTER,
         "tsfattr enter: fd %d fP 0x%lX name %s f_flags 0x%X flags 0x%X "
         "iP 0x%lX command %d argP 0x%lX rCodeP 0x%lX",
         fileDesc, fP, fP->f_dentry->d_name.name, fP->f_flags, flags, iP, 
         command, argP, rCodeP);

  switch (iP->i_mode & S_IFMT)
  {
    case S_IFREG:
    case S_IFDIR:
      vinfoP = (struct MMFSVInfo *)fP->private_data;
      break;

    case S_IFLNK:
    case S_IFBLK:
    case S_IFCHR:
    case S_IFIFO:
    case S_IFSOCK:
      /* At the moment we do have have any file ops enabled for the
       * special files on Linux, and thus we don't have an OpenFile object 
       * or vinfoP stored in fP.  This means that a special file can't
       * have e.g. ACLs associated with it.  Fixing this properly will 
       * require a redesign of our LFS interface.  For now, just fix the 
       * kernel panic caused by invalid vinfoP. */
      vinfoP = NULL;
      if ((command != REMOTE_STAT) &&   // UIDREMAP
          (command != GET_XATTR) &&     // GPFS_LINUX
          (command != SET_XATTR))       // GPFS_LINUX
      {
        code = 5;
        rc = ENOSYS;
        goto xerror;
      }
      break;
  }

  if (!GPFS_TYPE(iP))
  {
    code = 3;
    rc = EINVAL;
    /* If this file is not in GPFS filesystem, fill in the error reason in the 
     * argument structure's errReason field, which will be return to user's 
     * stderr finally. */   
    hdrp = (gpfsFcntlHeader_t *)argP;
    gncp = (genericStruct_t *)(hdrp + 1);
    if (gncp->structLen + sizeof(gpfsFcntlHeader_t) <= hdrp->totalLength)
    { 
       if(gncp->structType == GPFS_FCNTL_SET_REPLICATION)
          ((gpfsSetReplication_t *)gncp)->errReason = GPFS_FCNTL_ERR_NOT_GPFS_FILE;
       if(gncp->structType == GPFS_FCNTL_SET_STORAGEPOOL)
          ((gpfsSetStoragePool_t *)gncp)->errReason = GPFS_FCNTL_ERR_NOT_GPFS_FILE;
       if(gncp->structType == GPFS_FCNTL_RESTRIPE_DATA)
          ((gpfsRestripeData_t *)gncp)->errReason = GPFS_FCNTL_ERR_NOT_GPFS_FILE;
       if(gncp->structType == GPFS_FCNTL_RESTRIPE_RANGE)
          ((gpfsRestripeRange_t *)gncp)->errReason = GPFS_FCNTL_ERR_NOT_GPFS_FILE;
    }
    goto xerror;
  }

  privVfsP = VP_TO_PVP(iP);
  if (privVfsP == NULL)
  {
    code = 4;
    rc = ENOENT;
    goto xerror;
  }
  cnP = VP_TO_CNP(iP);

  rc = getCred(&eCred, &eCredP);
  if (rc)
  {
    code = 6;
    goto xerror;
  }

  rc = gpfs_ops.gpfsFattr(privVfsP, cnP, vinfoP, flags, command, argP,
                          rCodeP, eCredP);

xerror:
  putCred(&eCred, &eCredP);

  if (releFile)
    fput(fP);

  TRACE2(TRACE_VNODE, 1, TRCID_LINUXOPS_TSFATTR_EXIT,
         "tsfattr exit: code %d rc %d\n", code, rc);

  VFS_STAT_STOP;
  EXIT(0);
  return -rc;
}

int
tsfsattr(int command, void *argP)
{
  int rc;

  VFS_STAT_START(tsfattrCall);
  ENTER(0);

  TRACE1(TRACE_VNODE, 1, TRCID_LINUXOPS_TSFSATTR_ENTER,
        "tsfsattr enter: command %s\n",tsfsattrcmd_id2name(command));

  rc = gpfs_ops.gpfsFsAttr(command, argP);

  TRACE1(TRACE_VNODE, 1, TRCID_LINUXOPS_TSFSATTR_EXIT,
         "tsfsattr exit: rc %d\n", rc);

  VFS_STAT_STOP;
  EXIT(0);
  return -rc;
}

int
tsattr(const char *pathname, int command, void *argP,
       tsfattrReasonCodeInfo *rCodeP)
{
  struct gpfsVfsData_t *privVfsP;
  cxiNode_t *cnP;
  int code = 0;
  int rc = 0;
  int reason = 0;       /* local reason code for resetting caller's */
  ext_cred_t eCred;
  ext_cred_t *eCredP = &eCred;
  struct nameidata nd;
  struct inode *iP;
  Boolean relePath = false;


  /* if rCodeP was specified then immediately reset reason field */
  ENTER(0);
  if (rCodeP != NULL)
  {
    rc = cxiCopyOut((char*)&reason, (char*)&(rCodeP->reason), sizeof(reason));
    if (rc != 0)
    {
      code = 1;
      goto xerror;
    }
  }

  rc = USER_PATH(pathname, &nd);
  if (rc)
  {
    rc = -rc;
    code = 2;
    goto xerror;
  }
  relePath = true;

  iP = NDP_TO_IP(&nd);
  DBGASSERT(iP != NULL);

  TRACE4(TRACE_VNODE, 1, TRCID_LINUXOPS_TSATTR_ENTER,
         "tsattr enter: iP 0x%lX command %d argP 0x%lX rCodeP 0x%lX\n",
         iP, command, argP, rCodeP);

  if (!GPFS_TYPE(iP))
  {
    code = 3;
    rc = EINVAL;
    goto xerror;
  }

  privVfsP = VP_TO_PVP(iP);
  if (privVfsP == NULL)
  {
    code = 4;
    rc = ENOENT;
    goto xerror;
  }
  cnP = VP_TO_CNP(iP);

  rc = getCred(&eCred, &eCredP);
  if (rc)
  {
    code = 5;
    goto xerror;
  }

  rc = gpfs_ops.gpfsFattr(privVfsP, cnP, NULL, 0, command, argP, rCodeP,
                          eCredP);

xerror:
  putCred(&eCred, &eCredP);

  if (relePath)
    PATH_PUT(&nd);

  TRACE2(TRACE_VNODE, 1, TRCID_LINUXOPS_TSATTR_EXIT,
         "tsattr exit: code %d rc %d\n", code, rc);

  EXIT(0);
  return -rc;
}

int
kxGetACL(const char *pathname, int flags, void *aclP)
{
  int rc = 0;
  int code = 0;
  cxiNode_t *cnP;
  struct gpfsVfsData_t *privVfsP;
  ext_cred_t eCred;
  ext_cred_t *eCredP = &eCred;
  struct nameidata nd;
  struct inode *iP;
  Boolean relePath = false;

  /* Verify input exists */
  ENTER(0);
  if (pathname == NULL || aclP == NULL)
  {
    code = 1;
    rc = EINVAL;
    goto xerror;
  }

  rc = USER_PATH(pathname, &nd);
  if (rc)
  {
    rc = -rc;
    code = 2;
    goto xerror;
  }
  relePath = true;

  iP = NDP_TO_IP(&nd);
  DBGASSERT(iP != NULL);

  if (!GPFS_TYPE(iP))
  {
    code = 3;
    rc = EINVAL;
    goto xerror;
  }

  privVfsP = VP_TO_PVP(iP);
  if (privVfsP == NULL)
  {
    code = 4;
    rc = ENOENT;
    goto xerror;
  }

  cnP = VP_TO_CNP(iP);

  /* Retrieve the ACL data */
  rc = getCred(&eCred, &eCredP);
  if (rc)
  {
    code = 5;
    goto xerror;
  }

  rc = gpfs_ops.gpfsGetAcl(privVfsP, cnP, flags, aclP, eCredP);

xerror:
  putCred(&eCred, &eCredP);

  if (relePath)
    PATH_PUT(&nd);

  TRACE2(TRACE_VNODE, 1, TRCID_LINUXOPS_KXGETACL_EXIT,
         "kxGetACL exit: code %d rc %d\n", code, rc);
  EXIT(0);
  return -rc;
}

int
kxPutACL(const char *pathname, int flags, void *aclP)
{
  int rc = 0;
  int code = 0;
  cxiNode_t *cnP;
  struct gpfsVfsData_t *privVfsP;
  ext_cred_t eCred;
  ext_cred_t *eCredP = &eCred;
  struct nameidata nd;
  struct inode *iP;
  Boolean relePath = false;

  /* Verify input exists */
  ENTER(0);
  if (pathname == NULL || aclP == NULL)
  {
    code = 1;
    rc = EINVAL;
    goto xerror;
  }

  rc = USER_PATH(pathname, &nd);
  if (rc)
  {
    rc = -rc;
    code = 2;
    goto xerror;
  }
  relePath = true;

  iP = NDP_TO_IP(&nd);
  DBGASSERT(iP != NULL);

  if (!GPFS_TYPE(iP))
  {
    code = 3;
    rc = EINVAL;
    goto xerror;
  }

  privVfsP = VP_TO_PVP(iP);
  if (privVfsP == NULL)
  {
    code = 4;
    rc = ENOENT;
    goto xerror;
  }

  cnP = VP_TO_CNP(iP);

  /* Put the ACL data */
  rc = getCred(&eCred, &eCredP);
  if (rc)
  {
    code = 5;
    goto xerror;
  }

  rc = gpfs_ops.gpfsPutAcl(privVfsP, cnP, flags, aclP, eCredP);

xerror:
  putCred(&eCred, &eCredP);

  if (relePath)
    PATH_PUT(&nd);

  TRACE2(TRACE_VNODE, 1, TRCID_LINUXOPS_KXPUTACL_EXIT,
         "kxPutACL exit: code %d rc %d\n", code, rc);
  EXIT(0);
  return -rc;
}

/* aclP points struct ACL consists of header information (length, type, etc.)*/
int
vfs_getacl(struct dentry *dentry, void *aclP)
{
  int rc = 0;
  int code = 0;
  cxiNode_t *cnP;
  struct gpfsVfsData_t *privVfsP;
  ext_cred_t eCred;
  ext_cred_t *eCredP = &eCred;
  struct inode *iP;
  Boolean relePath = false;

  /* Verify input exists */
  ENTER(0);
  if (aclP == NULL)
  {
    code = 1;
    rc = EINVAL;
    goto xerror;
  }
  iP = dentry->d_inode;
  DBGASSERT(iP != NULL);

  if (!GPFS_TYPE(iP))
  {
    code = 3;
    rc = EINVAL;
    goto xerror;
  }
  privVfsP = VP_TO_PVP(iP);
  if (privVfsP == NULL)
  {
    code = 4;
    rc = ENOENT;
    goto xerror;
  }
  cnP = VP_TO_CNP(iP);

  /* Retrieve the ACL data */
  rc = getCred(&eCred, &eCredP);
  if (rc)
  {
    code = 5;
    goto xerror;
  }
  rc = gpfs_ops.gpfsGetAcl(privVfsP, cnP, GPFS_GETACL_STRUCT, aclP, eCredP);

xerror:
  putCred(&eCred, &eCredP);

  TRACE2(TRACE_VNODE, 1, TRCID_LINUXOPS_VFS_GETACL_EXIT,
         "vfs_getacl exit: code %d rc %d\n", code, rc);
  EXIT(0);
  return -rc;
}

int
vfs_setacl(struct dentry *dentry, void *aclP)
{
  int rc = 0;
  int code = 0;
  cxiNode_t *cnP;
  struct gpfsVfsData_t *privVfsP;
  ext_cred_t eCred;
  ext_cred_t *eCredP = &eCred;
  struct inode *iP;
  Boolean relePath = false;

  /* Verify input exists */
  ENTER(0);
  if (aclP == NULL)
  {
    code = 1;
    rc = EINVAL;
    goto xerror;
  }
  iP = dentry->d_inode;
  DBGASSERT(iP != NULL);

  if (!GPFS_TYPE(iP))
  {
    code = 3;
    rc = EINVAL;
    goto xerror;
  }
  privVfsP = VP_TO_PVP(iP);
  if (privVfsP == NULL)
  {
    code = 4;
    rc = ENOENT;
    goto xerror;
  }
  cnP = VP_TO_CNP(iP);

  /* Retrieve the ACL data */
  rc = getCred(&eCred, &eCredP);
  if (rc)
  {
    code = 5;
    goto xerror;
  }
  rc = gpfs_ops.gpfsPutAcl(privVfsP, cnP, GPFS_PUTACL_STRUCT, aclP, eCredP);

xerror:
  putCred(&eCred, &eCredP);

  TRACE2(TRACE_VNODE, 1, TRCID_LINUXOPS_VFS_SETACL_EXIT,
         "vfs_setacl exit: code %d rc %d\n", code, rc);
  EXIT(0);
  return -rc;
}


#if !defined(GPFS_ARCH_X86_64) && !defined(GPFS_ARCH_IA64)
/* madvise is handled as a kernel extension / ioctl call until support
   is provided by glibc */
#if LINUX_KERNEL_VERSION >= 2061900
/* needs a fix */
#else
static inline _syscall3(int,madvise,void*,addr,size_t,len,int,advice);
#endif
#endif

int
kxMadvise(void* addr, size_t len, int advice)
{
#if defined(GPFS_ARCH_X86_64) || defined(GPFS_ARCH_IA64)
  /* kxMadvise should never be called since glibc has madvice */
  return -ENOSYS;
#else
#if LINUX_KERNEL_VERSION >= 2061900
  /* needs a fix */
  return -ENOSYS;
#else
  return madvise(addr, len, advice);
#endif
#endif
}

/* Get the kernel thread ID. */
cxiThreadId
kxGetThreadID()
{
  return (cxiThreadId)current->pid;
}

#ifdef FILESETMAPPING_API
int kxGetFilesetId(const char *pathname, const char *name, int *idP)
{
  int rc = 0;
  int code = 0;
  struct nameidata nd;
  struct gpfsVfsData_t *privVfsP;
  ext_cred_t  eCred;
  ext_cred_t *eCredP = &eCred;
  struct inode *iP;
  Boolean relePath = false;

  ENTER(0);

  TRACE3(TRACE_VNODE, 2, TRCID_LINUXOPS_GETFILESETID_ENT,
         "kxGetFilesetId: enter pathname 0x%x, name 0x%x, euid %d\n",
          pathname, name, CRED(current, euid));

  /* Verify input exists */
  if (pathname == NULL || name == NULL)
  {
    code = 1;
    rc = EINVAL;
    goto xerror;
  }

  rc = USER_PATH(pathname, &nd);

  if (rc)
  {
    rc = -rc;
    code = 2;
    goto xerror;
  }

  relePath = true;

  iP = NDP_TO_IP(&nd);
  DBGASSERT(iP != NULL);

  if (!GPFS_TYPE(iP))
  {
    code = 3;
    rc = EINVAL;
    goto xerror;
  }

  privVfsP = VP_TO_PVP(iP);
  if (privVfsP == NULL)
  {
    code = 4;
    rc = ENOENT;
    goto xerror;
  }

  rc = getCred(&eCred, &eCredP);
  if (rc)
  {
    code = 5;
    goto xerror;
  }

  rc = gpfs_ops.gpfsGetFilesetId(privVfsP, name, idP, eCredP);

xerror:
  putCred(&eCred, &eCredP);

  if (relePath)
    PATH_PUT(&nd);

  TRACE2(TRACE_VNODE, 1, TRCID_LINUXOPS_GETFILESETID_EXIT,
         "kxGetFilesetId: exit: code %d, rc %d\n",
         code, rc);

  EXIT(0);

  return -rc;
}
#endif /* FILESETMAPPING_API */


int kxQuotactl(const char *pathname, int cmd, int qid, void *bufferP)
{
  int rc = 0;
  int code = 0;
  cxiNode_t *cnP;
  struct gpfsVfsData_t *privVfsP;
  ext_cred_t  eCred;
  ext_cred_t *eCredP = &eCred;
  struct nameidata nd;
  struct inode *iP;
  Boolean relePath = false;
  ulong len;

  ENTER(0);

  TRACE4(TRACE_VNODE, 2, TRCID_LINUXOPS_TSQUOTACTL_K,
         "kxQuotactl: enter cmd 0x%x, qid %d, pathname 0x%x, euid %d\n",
          cmd, qid, pathname, CRED(current, euid));

  /* Verify input exists */
  if (pathname == NULL)
  {
    code = 1; // EBUF_ERROR
    rc = EINVAL;
    goto xerror;
  }

  TRACE0(TRACE_VNODE, 2, TRCID_LINUXOPS_TSQUOTACTL_K1,
       "kxQuotactl at [1]\n");

  rc = USER_PATH(pathname, &nd);

  TRACE1(TRACE_VNODE, 2, TRCID_LINUXOPS_TSQUOTACTL_K2,
        "kxQuotactl at [2] user_path_walk rc %d\n",
         rc);

  if (rc)
  {
    rc = -rc;
    code = 3;
    goto xerror;
  }

  relePath = true;

  iP = NDP_TO_IP(&nd);
  DBGASSERT(iP != NULL);

  if (!GPFS_TYPE(iP))
  {
    code = 3;
    rc = EINVAL;
    goto xerror;
  }

  privVfsP = VP_TO_PVP(iP);
  if (privVfsP == NULL)
  {
    code = 4;
    rc = ENOENT;
    goto xerror;
  }

  cnP = VP_TO_CNP(iP);
 
  rc = getCred(&eCred, &eCredP);

  TRACE4(TRACE_VNODE, 2, TRCID_LINUXOPS_TSQUOTACTL_K6,
         "kxQuotactl: gpfs_ops.gpfsQuotactl, cmd 0x%x, qid %d, eCredP 0x%x, bufferP 0x%x\n",
         cmd, qid, eCredP, bufferP);

  if (rc)
  {
    code = 5;
    goto xerror;
  }

  rc = gpfs_ops.gpfsQuotactl(privVfsP, cmd, qid, bufferP, eCredP, cnP);

  TRACE1(TRACE_VNODE, 2, TRCID_LINUXOPS_TSQUOTACTL_K7,
         "kxQuotactl: gpfs_ops.gpfsQuotactl rc %d\n", rc);
xerror:
  putCred(&eCred, &eCredP);

  if (relePath)
    PATH_PUT(&nd);

  TRACE3(TRACE_VNODE, 1, TRCID_LINUXOPS_TSQUOTACTL_EXIT,
         "kxQuotactl exit: code %d rc %d, pathname 0x%x\n", 
         code, rc, pathname);

  EXIT(0);

  return -rc;
}

#ifdef SMB_LOCKS
/* Exchange control between Samba, GPFS */
int
setSMBOpenLockControl(void *kopP, void *vP, void *infoP, int command, int lockmode)
{
  struct inode *iP;
  struct gpfsVfsData_t *privVfsP;
  struct MMFSVInfo *vinfoP;
  struct file *fP = (struct file *)vP;
  int rc = EINVAL;

  ENTER(0);

  TRACE2(TRACE_VNODE, 9, TRCID_SMBOPENLOCKCONTROL_ENTERL,
         "setSMBOpenLockControl entry - command %d, mode 0x%x",
         command, lockmode);

  vinfoP = (struct MMFSVInfo *)fP->private_data;

  iP = fP->f_dentry->d_inode;
  DBGASSERT(iP != NULL);

  if (!GPFS_TYPE(iP))
  {
    rc = 0;
    goto xerror;
  }

  privVfsP = VP_TO_PVP(iP);

  rc = gpfs_ops.SMBOpenLockControl(kopP, command, lockmode, iP->i_ino, vinfoP,
                                   privVfsP);

xerror:

  TRACE1(TRACE_VNODE, 1, TRCID_SMBOPENLOCKCONTROL_EXITL,
         "setSMBOpLockControl exit: rc %d\n", rc);

  EXIT(0);
  return (rc);
}

#include <oplock.h>
#include <asm/uaccess.h>
int
setSMBOplock(void *vP, void *infoP, int accessWant, int oplockWant,
             void *breakArgP, int *oplockGrantedP, Boolean isAsync)
{
  struct file *filp = (struct file *)vP;
  struct gpfsVfsData_t *privVfsP;
  cxiNode_t *cnP;
  struct MMFSVInfo *vinfoP;
  struct file *fP;
  struct inode *iP;
  int reason = 0;
  int code = 0;
  int flags;
  int rc;
  Boolean releFile = false;

  /* get file pointer and inode */
  ENTER(0);
  fP = filp;
  if (!fP)
  {
    code = 2;
    rc = EBADF;
    goto xerror;
  }

  get_file(fP);

  releFile = true;

  vinfoP = (struct MMFSVInfo *)fP->private_data;

  iP = fP->f_dentry->d_inode;
  DBGASSERT(iP != NULL);

  TRACE5(TRACE_VNODE, 1, TRCID_LINUXOPS_SETSMBOPLOCK_E,
         "setSMBOplock enter: fP 0x%lX acc %d oplock %d argP 0x%lX async %d\n",
         fP, accessWant, oplockWant, breakArgP, isAsync);

  if (!GPFS_TYPE(iP))
  {
    code = 3;
    rc = EINVAL;
    goto xerror;
  }

  privVfsP = VP_TO_PVP(iP);
  if (privVfsP == NULL)
  {
    code = 4;
    rc = ENOENT;
    goto xerror;
  }
  cnP = VP_TO_CNP(iP);

  rc = gpfs_ops.gpfsSetSMBOplock(privVfsP, cnP, vinfoP, accessWant, oplockWant,
                                 breakArgP, oplockGrantedP, isAsync);

xerror:
  if (releFile)
    fput(fP);

  TRACE2(TRACE_VNODE, 1, TRCID_LINUXOPS_SETSMBOPLOCK_X,
         "setSMBOplock exit: granted %d rc %d", *oplockGrantedP, rc);

  EXIT(0);
  return rc;
}

/* Share/Delegate interface for Samba */
int
kxGetShare(int fd, unsigned int share, unsigned int deny)
{
  int rc;
  struct file *fP = NULL;
  struct inode *iP;
  struct MMFSVInfo *vinfoP;
  struct gpfsVfsData_t *privVfsP;
  int shareWant;

  ENTER(0);

  fP = fget(fd);
  if (!fP || !fP->f_dentry || !fP->f_dentry->d_inode)
  {
    rc = EINVAL;
    goto xerror;
  }

  vinfoP = (struct MMFSVInfo *)fP->private_data;

  iP = fP->f_dentry->d_inode;
  if (!iP || !(privVfsP = VP_TO_PVP(iP)))
  {
    rc = EBADF;
    goto xerror;
  }

  if (!GPFS_TYPE(iP))
  {
    rc = 0;
    goto xerror;
  }

  TRACE6(TRACE_VNODE, 1, TRCID_LINUX_SHARE_ENTER,
         "kxGetShare: enter fd %d fP 0x%lX inode %lld infoP 0x%lX access 0x%X "
         "deny 0x%X\n", fd, fP, iP->i_ino, vinfoP, share, deny);

  /* Translate (and validate) the Samba share/deny arguments to our shareWant */

  rc = EINVAL; /* setup for XLATE_SAMBA_ACCESS/DENY */

  XLATE_SAMBA_ACCESS(share, shareWant);
  XLATE_SAMBA_DENY(deny, shareWant);

  /* Call to make the reservation */
  if ((share != 0) || (deny != 0))
    rc = gpfs_ops.gpfsReserveShare(fP, vinfoP, privVfsP, 0, shareWant, 
				   NULL, NULL, NULL);
  else
    rc = 0;

  /* give back any pre-existing rights */
  if ((rc == 0) && (0 != (coMask & ~shareWant)))
    rc = gpfs_ops.gpfsReserveShare(fP, vinfoP, privVfsP, RESERVE_DOWNGRADE, 
                                   coMask & ~shareWant, NULL, NULL, NULL);
xerror:
  if (fP)
    fput(fP);

  TRACE1(TRACE_VNODE, 1, TRCID_LINUX_SHARE_EXIT, 
         "kxGetShare: exit rc %d\n", rc);

  EXIT(0);
  return -rc;
}

int
kxGetDelegation(int fd, unsigned int delegate_access,
                void *cb_token, void *cookie)
{
  int rc = 0;
  int code = 0;
  int flags = 0;
  struct file *fP = NULL;
  struct inode *iP;
  struct MMFSVInfo *vinfoP;
  struct gpfsVfsData_t *privVfsP;
  int oplockWant;

  ENTER(0);

#ifdef CLUSTER_LEASES
  // Linux fcntl_setlease will call the fs to get the lease
  goto xerror;
#endif

#if 0
  /* As long as Samba write leases are not as strict as NFSv4 write
     delegations (are not required to be revoked on stat) we can
     allow them.
  */
  if (delegate_access == CXI_DELEGATE_WRITE)
  {
    rc = EOPNOTSUPP;
    goto xerror;
  }
#endif

  fP = fget(fd);
  if (!fP || !fP->f_dentry || !fP->f_dentry->d_inode)
  {
    rc = EINVAL;
    goto xerror;
  }

  vinfoP = (struct MMFSVInfo *)fP->private_data;

  iP = fP->f_dentry->d_inode;
  if (!iP || !(privVfsP = VP_TO_PVP(iP)))
  {
    rc = EBADF;
    goto xerror;
  }

  if (!GPFS_TYPE(iP))
  {
    rc = 0;
    goto xerror;
  }

  TRACE5(TRACE_VNODE, 1, TRCID_DELEGATE_ENTER,
         "kxGetDelegation: enter fd %d fP 0x%lX inode %lld vinfoP 0x%lX "
         "delegate 0x%x\n", fd, fP, iP->i_ino, vinfoP, delegate_access);

  /* Translate (and validate) the Samba delegate argument to our oplock flags */

  /* setup for XLATE_SAMBA_DELEGATE */
  rc = EINVAL;
  oplockWant = 0;

  if (delegate_access == CXI_DELEGATE_NONE)
    flags = RESERVE_DOWNGRADE;

  XLATE_SAMBA_DELEGATE(delegate_access, oplockWant);

  rc = gpfs_ops.gpfsReserveDelegation(fP, vinfoP, privVfsP, oplockWant, 
                                      flags, iP, NULL, NULL);

xerror:
  if (fP)
    fput(fP);

  TRACE1(TRACE_VNODE, 1, TRCID_DELEGATE_EXIT,
         "kxGetDelegation: exit rc %d\n", rc);

  EXIT(0);
  return -rc;
}
#endif

int kxWinOps(int fd, const char * pathname, int op, int flags, void *genArg1)
{
  int rc = 0;
  int code = 0;
  int crc;
  struct file *fP = NULL;
  struct inode *iP = NULL;
  struct nameidata nd;
  /* struct MMFSVInfo *vinfoP; */
  struct gpfsVfsData_t *privVfsP = NULL;
  cxiNode_t *cnP = NULL;
  ext_cred_t  eCred;
  ext_cred_t *eCredP = &eCred;
  cxiWinBasicInfo_t info;
  Boolean relePath = false;

  ENTER(0);

  if ((!genArg1) &&
      ((op == WIN_OP_GETATTRS) || 
       (op == WIN_OP_SETATTRS) ||
       (op == WIN_OP_REGISTER_CIFS_BUF)))
  {
    code = 2;
    rc = EINVAL;
    goto xerror;
  }

  if ((op == WIN_OP_GETATTRS) || (op == WIN_OP_SETATTRS))
  {
    if (pathname)
    {
      rc = USER_PATH(pathname, &nd);
      if (rc)
      {
        rc = -rc;
        code = 3;
        goto xerror;
      }
      relePath = true;

      iP = NDP_TO_IP(&nd);
      DBGASSERT(iP != NULL);
    }
    else
    {
      fP = fget(fd);
      if (!fP || !fP->f_dentry || !fP->f_dentry->d_inode)
      {
        code = 4;
        rc = EINVAL;
        goto xerror;
      }

      /* vinfoP = (struct MMFSVInfo *)fP->private_data; */

      iP = fP->f_dentry->d_inode;
    }

    if (!iP || !(privVfsP = VP_TO_PVP(iP)))
    {
      code = 5;
      rc = EBADF;
      goto xerror;
    }

    if (!GPFS_TYPE(iP))
    {
      code = 6;
      rc = EBADF;
      goto xerror;
    }

    cnP = VP_TO_CNP(iP);
  }

  rc = getCred(&eCred, &eCredP);
  if (rc)
    goto xerror;

  TRACE6(TRACE_VNODE, 1, TRCID_LINUX_SAMBA_OP_ENTER,
         "kxWinOps: enter op %d fd %d fP 0x%lX inode %lld flags 0x%X "
         "arg1 0x%X\n", op, fd, fP, iP? iP->i_ino: 0, flags, genArg1);

  rc = gpfs_ops.gpfsWinOps(op, flags, genArg1, privVfsP, cnP, &info, eCredP);

xerror:
  putCred(&eCred, &eCredP);

  if (relePath)
    PATH_PUT(&nd);

  if (fP)
    fput(fP);

  TRACE2(TRACE_VNODE, 1, TRCID_LINUX_SAMBA_OP_EXIT, 
         "kxWinOps: exit rc %d code %d\n", rc, code);

  EXIT(0);
  return -rc;
}

/* called from umount helper before calling umount system call  */

#define UMOUNT_BEGIN 0
#define UMOUNT_END 1
int kxUMount(int type, char *pathname, int force, int *ret_err)
{
    char kxnode[CXI_MAXPATHLEN+1];
    int rc = 0;
    int dmrc = 0;
    int code = 0;
    size_t len;
    struct cxiNode_t  *cnP = NULL;
    struct gpfsVfsData_t *privVfsP = NULL;
    void       *ndP = NULL;
    Boolean doDMEvents = false;

    ENTER(0);
    rc = cxiCopyInstr(pathname, kxnode, CXI_MAXPATHLEN, &len);
    LOGASSERT(rc == 0);
    TRACE4(TRACE_VNODE, 1, TRCID_KXUMOUNT_ENT,
         "kxUMount: enter type %d path %s force %d  ret_err %d\n",
          type,kxnode,force,*ret_err);
    
#ifdef DMAPI
    rc = cxiPathToVfsP((void **)&privVfsP,pathname, &ndP, (void **)&cnP, true);
    if (rc)
    {
      code = 1;
      goto exit;
    }         
    
    /* Generate PREUNMOUNT event for normal mounts. For force unmount
           it is generated from gpfs_s_umount_begin.
           umount helper calls with UMOUNT_BEGIN before calling umount
           system call.
    */

    if ( type == UMOUNT_BEGIN && !force)
    {
       dmrc = gpfs_ops.gpfsDmUnmountEvent(true,force,privVfsP,cnP,
                                        &doDMEvents, NULL, NULL, NULL, 0);
       if (doDMEvents )
         *ret_err = 1;
       else
         *ret_err = 0;
    }

        /* 
          umount helper calls with UMOUNT_END if umount system call fails.
    */

    if ( type == UMOUNT_END )
       dmrc = gpfs_ops.gpfsDmUnmountEvent(false,force,privVfsP,cnP,
                                     &doDMEvents, NULL, NULL, NULL, *ret_err);

#endif
exit:
    if (ndP)
      cxiPathRel(ndP);

    TRACE4(TRACE_VNODE, 1, TRCID_KXUMOUNT_EXT,
         "kxUMount: exit dmrc %d  ret_err %d rc %d code %d\n",
          dmrc,*ret_err,rc,code);
     EXIT(0);
     return rc;
}

int kxGetRealFileName(int fd, const char *pathname, char * realFileNameP, int *lenP)
{
  int rc = 0;
  struct file *fP = NULL;
  struct inode *diP = NULL;
  struct inode *iP = NULL;
  struct dentry *dentryP = NULL;
  cxiNode_t *cnDirP = NULL;
  cxiNode_t *cnFileP = NULL;
  struct gpfsVfsData_t *privVfsP;
  int nameLength = 0;
  int inputLen = 0;
  ext_cred_t  eCred;
  ext_cred_t *eCredP = &eCred;
  int code = 0;
  size_t len = 0;
  struct nameidata nd;
  Boolean relePath = false;
  char *dirPathName = NULL;
  char *filenameP = NULL;
  char *cP = NULL;
  char *curNameP = NULL;
  char *bufP = NULL;
  Boolean isCaseSensitive = false;
  int supportSambaLookup = false;

  /* Check if we are configured for case-insensitive samba lookkups. */
  supportSambaLookup = gpfs_ops.gpfsSupportSambaLookup();
  if (supportSambaLookup < 0)
  {
    rc = -(supportSambaLookup);
    code = 16;
    goto xerror;
  }

  /* Normally, we do case-insensitive lookups.
     The exception is for samba (when properly configured). */

  isCaseSensitive = (!cxiIsSambaThread() || !supportSambaLookup);

  bufP = (char *)cxiMallocUnpinned(CXI_MAXPATHLEN+1);
  if (bufP == NULL)
  {
    /* Unable to allocate kernel space for the name */
    code = 1;
    rc = ENOMEM;
    goto xerror;
  }
  curNameP = bufP;
  cxiMemset(curNameP, 0, CXI_MAXPATHLEN+1);

  ENTER(0);
  rc = getCred(&eCred, &eCredP);
  if (rc)
  {
    code = 15;
    goto xerror;
  }

  rc = cxiCopyIn((char *)lenP, (char *)&inputLen, sizeof(int));
  if (rc)
  {
    code = 3;
    goto xerror;
  }

  /* Pathnamne was given, so lookup the file without an open. */
  if (pathname)
  {
    rc = cxiCopyInstr(pathname, curNameP, CXI_MAXPATHLEN, &len);
    if (rc != 0)
    {
      code = 4;
      goto xerror;
    }

    nameLength = (int)len;

    TRACE2(TRACE_VNODE, 1, TRCIDGETREALNAME1,
           "kxGetRealFileName: file %s buffer len = %d", curNameP, inputLen);

    /* Check to see if length of user provided buffer is large enogh to hold actual name */
    if (len  > inputLen)
    {
      code = 5;
      rc = ENOSPC;
      goto xerror;
    }

    /* set filenameP to the last component of the path (the file), leaving
       keepng dirPathName pointing to the parent directory. */

    dirPathName = curNameP;
    cP = cxiStrchr(curNameP, '/');
    if (!cP)
    {
      code = 6;
      rc = EINVAL;
      goto xerror;
    }

    while (cP)
    {
      filenameP = cP;
      curNameP ++;
      cP = cxiStrchr(curNameP, '/');
    }
    *filenameP = '\0';

    /* Previous path components must have previously been translated, so we
       can lookup the parent directory name. */

    rc = PATH_LOOKUP(dirPathName, LOOKUP_FOLLOW | LOOKUP_DIRECTORY, &nd);
    if (rc)
    {
      /* Parent directory not valid */
      rc = -rc;
      code = 7;
      goto xerror;
    }
    relePath = true;

    diP = NDP_TO_IP(&nd);
    if (!diP || !(privVfsP = VP_TO_PVP(diP)))
    {
      code = 8;
      rc = EBADF;
      goto xerror;
    }

    if (!GPFS_TYPE(diP))
    {
      /* Parent directory is not GPFS */
      code = 9;
      rc = EBADF;
      goto xerror;
    }

    cnDirP = VP_TO_CNP(diP);

    if (!cnDirP)
    {
      code = 10;
      rc = EBADF;
      goto xerror;
    }

    nameLength = len -(cxiStrlen(dirPathName)); /* remaing in buffer (filename only) */

    *filenameP = '/';
    filenameP++;

    rc = gpfs_ops.gpfsLookupRealName(cnDirP, filenameP, privVfsP, eCredP, filenameP, &nameLength);

#if 1
    /* gpfsLookupRealName returned the translated filename on top of the input
       name part of the full pathname and nameLength of this name part only.
       Return the full translated pathname and its length to our caller. */
    nameLength = (int)len;
    curNameP = dirPathName;
#else
    /* Return the filename only (not the entire path) and its length. */
    curNameP = filenameP;
#endif
  }

  else /* file descriptor case */
  {
   fP = fget(fd);
    if (!fP || !fP->f_dentry || !fP->f_dentry->d_inode ||
        !fP->f_dentry->d_parent || !fP->f_dentry->d_parent->d_inode )
    {
      code = 11;
      rc = EINVAL;
      goto xerror;
    }

    iP = fP->f_dentry->d_inode;
    /* This dentry will have an hold so not calling dget on this */
    dentryP = fP->f_dentry;
    diP = fP->f_dentry->d_parent->d_inode;

    if (!GPFS_TYPE(iP))
    {
      code = 12;
      rc = EBADF;
      goto xerror;
    }
  
    if (!iP || !(privVfsP = VP_TO_PVP(iP)))
    {
      rc = EBADF;
      goto xerror;
    }

    cnDirP  = VP_TO_CNP(diP);
    cnFileP = VP_TO_CNP(iP);
    if (!cnDirP || !cnFileP)
    {
      code = 13;
      rc = EBADF;
      goto xerror;
    }

    nameLength = cxiStrlen(dentryP->d_name.name) + 1;

    /* Check to see if length of user provided buffer is large enogh to hold actual name */
    if (nameLength > inputLen)
    {
      code = 14;
      rc = ENOSPC;
      goto xerror;
    }

    TRACE2(TRACE_VNODE, 1, TRCIDGETREALNAME2,
           "kxGetRealFileName:  dentryName: %s, namelen = %d\n",
           dentryP->d_name.name, nameLength);


    /* If this is a root dir get its name from dentry itself. No need to perform
       lookup */
    if (iP && iP->i_ino == INODENUM_ROOTDIR_FILE)
    {
      cxiStrcpy(curNameP, dentryP->d_name.name);
      goto xerror;
    }
    /* SAMBA can force GPFS on Linux to do a caseinsensitive lookup.
     * In that case, name of file stored in dcache may not be actual filename.
     * To get real file name, pass on the FileUID so that after name lookup
     * FileUID can be matched to get real file name. This lookup does not create
     * negative dcache entry if name not found. It also does not perform
     * relookup in case rename has happened on other node. Since fget has already
     * an hold on dentry so we can be sure of it not going away.
     */
  
    rc = gpfs_ops.gpfsLookupFileWithFileId (cnDirP,
                                            dentryP->d_name.name,
                                            isCaseSensitive,
                                            privVfsP,
                                            eCredP,
                                            cnFileP,
                                            curNameP,
                                            &nameLength);
    }

xerror:
  putCred(&eCred, &eCredP);

  if (rc == 0)
    cxiCopyOut(curNameP, realFileNameP, nameLength+1);

  if (rc == 0 || rc == ENOSPC)
    cxiCopyOut((char *)&nameLength, (char *)lenP, sizeof(int));

  if (relePath)
    PATH_PUT(&nd);

  if (fP)
    fput(fP);

  if (bufP)
    cxiFreeUnpinned(bufP);

  TRACE3(TRACE_VNODE, 1, TRCIDGETREALNAME3,
         "kxGetRealFileName:  actualnamelen= %d, rc= %d, code=%d\n",
          nameLength, rc, code);

  EXIT(0);
  return -rc;
}

#ifdef CLONE_FILE
/* Create a clone copy of a file */
int kxCloneFile(CloneOp clop, const char *sourcePathP, const char *destPathP)
{
  int rc = 0;
  int code = 0;
  struct gpfsVfsData_t *privVfsP;
  struct nameidata sourceNd, destNd;
  struct inode *sourceIP, *destIP;
  cxiNode_t *cnP, *cnDirP = NULL;
  char *bufP = NULL, *nameP = NULL;
  const char *dirNameP;
  size_t len = 0;
  mm_segment_t oldfs;
  ext_cred_t eCred;
  ext_cred_t *eCredP = &eCred;
  Boolean relePath = false, relePath2 = false;

  ENTER(0);

  rc = USER_PATH(sourcePathP, &sourceNd);
  if (rc != 0)
  {
    rc = -rc;
    code = 2;
    goto xerror;
  }
  relePath = true;

  sourceIP = NDP_TO_IP(&sourceNd);
  DBGASSERT(sourceIP != NULL);

  if (!GPFS_TYPE(sourceIP))
  {
    code = 3;
    rc = EINVAL;
    goto xerror;
  }

  privVfsP = VP_TO_PVP(sourceIP);
  if (privVfsP == NULL)
  {
    code = 4;
    rc = ENOENT;
    goto xerror;
  }
  cnP = VP_TO_CNP(sourceIP);

  rc = getCred(&eCred, &eCredP);
  if (rc != 0)
  {
    code = 5;
    goto xerror;
  }

  if (clop == coCopy || destPathP != NULL)
  {
    /* Split the destination file name from directory */
    bufP = (char *)cxiMallocUnpinned(CXI_MAXPATHLEN+1);
    if (bufP == NULL)
    {
      code = 6;
      rc = ENOMEM;
      goto xerror;
    }
    rc = cxiCopyInstr(destPathP, bufP, CXI_MAXPATHLEN, &len);
    if (rc != 0)
    {
      rc = -rc;
      code = 7;
      goto xerror;
    }
    if (len == 0)
    {
      code = 8;
      rc = EINVAL;
      goto xerror;
    }
    if (bufP[len-1] == '/')
    {
      code = 9;
      rc = EINVAL;
      goto xerror;
    }
    nameP = cxiStrrchr(bufP, '/');
    if (nameP == NULL)
    {
      nameP = bufP;
      dirNameP = ".";
    }
    else
    {
      *nameP++ = '\0';
      dirNameP = bufP;
    }

    /* Look up the parent directory*/
    oldfs = get_fs();
    set_fs(get_ds());
    rc = USER_PATH(dirNameP, &destNd);
    set_fs(oldfs);
    if (rc != 0)
    {
      rc = -rc;
      code = 10;
      goto xerror;
    }
    relePath2 = true;

    destIP = NDP_TO_IP(&destNd);
    DBGASSERT(destIP != NULL);

    if (destIP->i_op->lookup == NULL)
    {
      code = 11;
      rc = ENOTDIR;
      goto xerror;
    }
    if (!GPFS_TYPE(destIP))
    {
      code = 12;
      rc = EINVAL;
      goto xerror;
    }
    if (VP_TO_PVP(destIP) != privVfsP)
    {
      code = 13;
      rc = EXDEV;
      goto xerror;
    }
    cnDirP = VP_TO_CNP(destIP);
  }
  rc = gpfs_ops.gpfsCloneFile(privVfsP, clop, cnP, cnDirP, NULL, nameP,
                              IS_CASE_SENSITIVE(), current->fs->umask,
                              eCredP);

xerror:
  putCred(&eCred, &eCredP);
  if (relePath)
    PATH_PUT(&sourceNd);
  if (relePath2)
    PATH_PUT(&destNd);
  if (bufP != NULL)
    cxiFreeUnpinned(bufP);

  TRACE2(TRACE_VNODE, 1, TRCID_KXCLONEFILE_EXIT, 
         "kxCloneFile exit: rc %d code %d", rc, code);

  EXIT(0);
  return -rc;
}

/* Copy blocks from clone parent to child, or split clone child from parent */
int kxDeclone(DecloneOp dcop, int fd, int ancLimit, UInt32 nBlocksHi,
              UInt32 nBlocksLo, Int64 *offsetP)
{
  int rc = 0;
  int code = 0;
  struct file *fP = NULL;
  ext_cred_t  eCred;
  ext_cred_t *eCredP = &eCred;
  struct dentry *dentryP = NULL;
  struct inode *iP = NULL;
  cxiNode_t *cnFileP = NULL;
  struct gpfsVfsData_t *privVfsP;
  Int64 nBlocks = Cast32To64(nBlocksHi, nBlocksLo), offset = 0;

  ENTER(0);

  rc = getCred(&eCred, &eCredP);
  if (rc != 0)
  {
    code = 1;
    goto xerror;
  }

  if (dcop == dcCopy && offsetP == NULL)
  {
    code = 2;
    rc = EINVAL;
    goto xerror;
  }

  if (offsetP != NULL)
  {
    rc = cxiCopyIn((char *)offsetP, (char *)&offset, sizeof(Int64));
    if (rc != 0)
    {
      code = 3;
      goto xerror;
    }
  }

  fP = fget(fd);
  if (fP == NULL || fP->f_dentry == NULL)
  {
    code = 4;
    rc = EBADF;
    goto xerror;
  }

  dentryP = fP->f_dentry;
  if (dentryP->d_inode == NULL)
  {
    code = 5;
    rc = ENOENT;
    goto xerror;
  }
  iP = dentryP->d_inode;

  if (!GPFS_TYPE(iP))
  {
    code = 6;
    rc = EBADF;
    goto xerror;
  }

  privVfsP = VP_TO_PVP(iP);
  if (privVfsP == NULL)
  {
    code = 7;
    rc = EBADF;
    goto xerror;
  }

  cnFileP = VP_TO_CNP(iP);
  if (cnFileP == NULL)
  {
    code = 8;
    rc = EBADF;
    goto xerror;
  }

  rc = gpfs_ops.gpfsDeclone(privVfsP, dcop, cnFileP, ancLimit, nBlocks,
                            NULL, eCredP, &offset);
  if (rc != 0)
  {
    code = 9;
    goto xerror;
  }

  if (offsetP != NULL)
  {
    rc = cxiCopyOut((char *)&offset, (char *)offsetP, sizeof(Int64));
    if (rc != 0)
    {
      code = 10;
      goto xerror;
    }
  }

xerror:
  putCred(&eCred, &eCredP);
  if (fP)
    fput(fP);

  TRACE3(TRACE_VNODE, 1, TRCID_DECLONE,
         "kxDeclone exit: fd %d rc %d code %d", fd, rc, code);
  EXIT(0);
  return -rc;
}
#endif /* CLONE_FILE */


#ifdef FFDEDUPE

static IntRC fgetValidate(int fd, struct file **fPP)
{
  int rc = 0;
  const char *traceInfo = "none";
  struct file *fP = NULL;
  struct dentry *dentryP = NULL;
  struct inode *iP = NULL;
  struct inode *diP = NULL;

  *fPP = NULL;

  fP = fget(fd);
  if (!fP || !fP->f_dentry || !fP->f_dentry->d_parent )
  {
    traceInfo = "fget";
    rc = EBADF;
    goto xerror;
  }

  dentryP = fP->f_dentry;
  iP = dentryP->d_inode;
  diP = dentryP->d_parent->d_inode;
  if (!iP || !diP)
  {
    traceInfo = "d_inode";
    rc = ENOENT;
    goto xerror;
  }

  if (!GPFS_TYPE(iP))
  {
    traceInfo = "GPFS_TYPE";
    rc = EBADF;
    goto xerror;
  }

  if (!VP_TO_PVP(iP))
  {
    traceInfo = "VP_TO_PVP";
    rc = EBADF;
    goto xerror;
  }

  if (!VP_TO_CNP(iP))
  {
    traceInfo = "VP_TO_CNP";
    rc = EBADF;
    goto xerror;
  }

  *fPP = fP;

xerror:
  if (rc)
  {
    if (fP != NULL)
      fput(fP);

    TRACE3(TRACE_VNODE, 1, TRCID_VALIDATEFILE,
           "fgetValidate error: fd %d rc %d info: %s",
           fd, rc, traceInfo);
  }

  return rc;
}


IntRC kxCloneFattach(int childDesc, int oldParentDesc, int newParentDesc)
{
  int rc = 0;
  const char *traceInfo = "none";
  ext_cred_t  eCred;
  ext_cred_t *eCredP = &eCred;
  struct file *childFileP = NULL;
  struct file *oldParentFileP = NULL;
  struct file *newParentFileP = NULL;
  struct inode *childInodeP = NULL;
  struct inode *oldParentInodeP = NULL;
  struct inode *newParentInodeP = NULL;
  struct gpfsVfsData_t *privVfsP;

  ENTER(0);

  TRACE3(TRACE_VNODE, 1, TRCID_KXCLONEFATTACH_ENTER,
         "kxCloneFattach enter: childDesc %d oldParentDesc %d newParentDesc %d",
         childDesc, oldParentDesc, newParentDesc);

  rc = getCred(&eCred, &eCredP);
  if (rc)
  {
    traceInfo = "getCred";
    goto xerror;
  }

  rc = fgetValidate(childDesc, &childFileP);
  if (rc)
  {
    traceInfo = "fgetValidate(childDesc)";
    goto xerror;
  }
  rc = fgetValidate(oldParentDesc, &oldParentFileP);
  if (rc)
  {
    traceInfo = "fgetValidate(oldParentDesc)";
    goto xerror;
  }
  rc = fgetValidate(newParentDesc, &newParentFileP);
  if (rc)
  {
    traceInfo = "fgetValidate(newParentDesc)";
    goto xerror;
  }

  childInodeP = childFileP->f_dentry->d_inode;
  privVfsP = VP_TO_PVP(childInodeP);

  oldParentInodeP = oldParentFileP->f_dentry->d_inode;
  if (privVfsP != VP_TO_PVP(oldParentInodeP))
  {
    traceInfo = "oldParent VFS mismatch";
    rc = EINVAL;
    goto xerror;
  }

  newParentInodeP = newParentFileP->f_dentry->d_inode;
  if (privVfsP != VP_TO_PVP(newParentInodeP))
  {
    traceInfo = "newParent VFS mismatch";
    rc = EINVAL;
    goto xerror;
  }

  /* FFDEDUPE FIXME -- do we need INODE_LOCK(iP)? */

  rc = gpfs_ops.gpfsCloneAttach(privVfsP,
                                VP_TO_CNP(childInodeP), 
                                VP_TO_CNP(oldParentInodeP), 
                                VP_TO_CNP(newParentInodeP), 
                                eCredP);
  if (rc != 0)
  {
    traceInfo = "gpfs_ops.gpfsCloneAttach";
    goto xerror;
  }

xerror:
  putCred(&eCred, &eCredP);

  if (childFileP)
    fput(childFileP);
  if (oldParentFileP)
    fput(oldParentFileP);
  if (newParentFileP)
    fput(newParentFileP);

  TRACE5(TRACE_VNODE, 1, TRCID_KXCLONEFATTACH_EXIT,
         "kxCloneFattach exit: childDesc %d oldParentDesc %d newParentDesc %d rc %d info: %s",
         childDesc, oldParentDesc, newParentDesc, rc, traceInfo);

  EXIT(0);
  return -rc;
}


IntRC kxFhashattr(int fileDesc,
                  const gpfs_hashattr_t *compareAttrP, 
                  const gpfs_hashattr_t *newAttrP,
                  gpfs_hashattr_t *oldAttrP)
{
  int rc = 0;
  const char *traceInfo = "none";
  ext_cred_t  eCred;
  ext_cred_t *eCredP = &eCred;
  gpfs_hashattr_t compareAttr;
  gpfs_hashattr_t newAttr;
  gpfs_hashattr_t oldAttr;
  struct file *fP = NULL;
  struct inode *iP = NULL;

  ENTER(0);

  TRACE3(TRACE_VNODE, 1, TRCID_KXHASHATTR_ENTER,
         "kxFhashattr enter: fileDesc %d compareAttr %016llx... newAttr %016llx...",
         fileDesc, *(BigEndUInt64 *)compareAttrP, *(BigEndUInt64 *)newAttrP);

  rc = getCred(&eCred, &eCredP);
  if (rc)
  {
    traceInfo = "getCred";
    goto xerror;
  }

  rc = cxiCopyIn((char *)compareAttrP, (char *)&compareAttr, sizeof(gpfs_hashattr_t));
  if (rc)
  {
    traceInfo = "cxiCopyIn(compareAttrP)";
    goto xerror;
  }

  rc = cxiCopyIn((char *)newAttrP, (char *)&newAttr, sizeof(gpfs_hashattr_t));
  if (rc)
  {
    traceInfo = "cxiCopyIn(newAttrP)";
    goto xerror;
  }

  rc = cxiCopyIn((char *)oldAttrP, (char *)&oldAttr, sizeof(gpfs_hashattr_t));
  if (rc)
  {
    traceInfo = "cxiCopyIn(oldAttrP)";
    goto xerror;
  }

  rc = fgetValidate(fileDesc, &fP);
  if (rc)
  {
    traceInfo = "fgetValidate(fileDesc)";
    goto xerror;
  }

  iP = fP->f_dentry->d_inode;

  /* FFDEDUPE FIXME -- do we need INODE_LOCK(iP)? */

  rc = gpfs_ops.gpfsHashattr(VP_TO_PVP(iP), VP_TO_CNP(iP), &compareAttr, &newAttr, eCredP, &oldAttr);
  if (rc != 0)
  {
    traceInfo = "gpfs_ops.gpfsHashattr";
    goto xerror;
  }

  rc = cxiCopyOut((char *)&oldAttr, (char *)oldAttrP, sizeof(gpfs_hashattr_t));
  if (rc)
  {
    traceInfo = "cxiCopyOut(oldAttr)";
    goto xerror;
  }

xerror:
  putCred(&eCred, &eCredP);

  if (fP)
    fput(fP);

  TRACE4(TRACE_VNODE, 1, TRCID_KXHASHATTR_ERROR_EXIT,
         "kxFhashattr exit: fileDesc %d oldAttr %016llx... rc %d info: %s",
         fileDesc, *(BigEndUInt64 *)oldAttrP, rc, traceInfo);

  EXIT(0);
  return -rc;
}


IntRC kxFlink(int dirDesc,
              const char *name,
              int fileDesc,
              int flags)
{
  int rc = 0;
  const char *traceInfo = "none";
  ext_cred_t  eCred;
  ext_cred_t *eCredP = &eCred;
  char nameBuffer[CXI_MAXPATHLEN + 1];
  size_t nameLength = 0;
  struct qstr nameQstr;
  Boolean caseSensitive = true;

  struct file *dfP = NULL;
  struct inode *diP = NULL;
  struct dentry *dentryP = NULL;
  struct file *fP = NULL;
  struct inode *iP = NULL;

  ENTER(0);
  TRACE4(TRACE_VNODE, 1, TRCID_KXFLINK_ENTER,
         "kxFlink enter: dirDesc %d name %lx fileDesc %d flags 0x%X",
         dirDesc, name, fileDesc, flags);

  rc = getCred(&eCred, &eCredP);
  if (rc)
  {
    traceInfo = "getCred";
    goto xerror;
  }

  rc = cxiCopyInstr((char *)name, nameBuffer, sizeof(nameBuffer), &nameLength);
  if (rc != 0)
  {
    traceInfo = "cxiCopyInstr(name)";
    goto xerror;
  }

  rc = fgetValidate(dirDesc, &dfP);
  if (rc)
  {
    traceInfo = "fgetValidate(dirDesc)";
    goto xerror;
  }

  rc = fgetValidate(fileDesc, &fP);
  if (rc)
  {
    traceInfo = "fgetValidate(fileDesc)";
    goto xerror;
  }

  diP = dfP->f_dentry->d_inode;
  iP = fP->f_dentry->d_inode;

  nameQstr.hash = full_name_hash(nameBuffer, nameLength);
  nameQstr.len = nameLength;
  nameQstr.name = nameBuffer;
  dentryP = d_lookup(dfP->f_dentry, &nameQstr);

  rc = gpfs_ops.gpfsLink(VP_TO_PVP(diP), VP_TO_CNP(iP), VP_TO_CNP(diP), 
                         dentryP, nameBuffer, caseSensitive, false, eCredP);
  if (rc != 0)
  {
    traceInfo = "gpfs_ops.gpfsLink";
    goto xerror;
  }

  iP->i_sb->s_dirt = 1;

xerror:
  if (dentryP)
    dput(dentryP);
  if (fP)
    fput(fP);
  if (dfP)
    fput(dfP);
  putCred(&eCred, &eCredP);

  TRACE4(TRACE_VNODE, 1, TRCID_KXFLINK_EXIT,
         "kxFlink exit: dirDesc %d fileDesc %d rc %d info: %s",
         dirDesc, fileDesc, rc, traceInfo);
  EXIT(0);
  return -rc;
}


/* This function will unlink name from dirDesc. It first verifies that
   name points to the same file as fileDesc. fileDesc may have been opened
   with a different name -- only its inode and gen number are relevant. */

IntRC kxFunlink(int dirDesc,
                const char *name,
                int fileDesc,
                int flags)
{
  int rc = 0;
  const char *traceInfo = "none";
  char nameBuffer[CXI_MAXPATHLEN + 1];
  size_t nameLength = 0;
  Boolean caseSensitive = true;

  struct file *dfP = NULL;
  struct inode *diP = NULL;
  struct dentry *dentryP = NULL;
  struct file *fP = NULL;
  struct inode *iP = NULL;

  ENTER(0);
  TRACE4(TRACE_VNODE, 1, TRCID_KXFUNLINK_ENTER,
         "kxFlink enter: dirDesc %d name %lx fileDesc %d flags 0x%X",
         dirDesc, name, fileDesc, flags);

  rc = cxiCopyInstr((char *)name, nameBuffer, sizeof(nameBuffer), &nameLength);
  if (rc != 0)
  {
    traceInfo = "cxiCopyInstr(name)";
    goto xerror;
  }

  rc = fgetValidate(dirDesc, &dfP);
  if (rc)
  {
    traceInfo = "fgetValidate(dirDesc)";
    goto xerror;
  }

  rc = fgetValidate(fileDesc, &fP);
  if (rc)
  {
    traceInfo = "fgetValidate(fileDesc)";
    goto xerror;
  }

  diP = dfP->f_dentry->d_inode;
  INODE_LOCK(diP);

  dentryP = lookup_one_len(nameBuffer, dfP->f_dentry, nameLength);
  if (dentryP == NULL)
  {
    rc = ENOENT;
    traceInfo = "lookup_one_len";
    goto xerror;
  }

  iP = igrab(dentryP->d_inode);
  if (iP == NULL)
  {
    rc = ESTALE;
    traceInfo = "igrab";
    goto xerror;
  }
  INODE_LOCK(iP);

  if (iP->i_ino != fP->f_dentry->d_inode->i_ino ||
      iP->i_generation != fP->f_dentry->d_inode->i_generation)
  {
    rc = ESTALE;
    traceInfo = "ino compare";
    goto xerror;
  }

  rc = gpfs_i_unlink(diP, dentryP);
  if (rc)
  {
    traceInfo = "gpfs_i_unlink";
    goto xerror;
  }

  d_drop(dentryP);

xerror:
  if (iP)
  {
    INODE_UNLOCK(iP);
    iput(iP);
  }
  if (dentryP)
    dput(dentryP);
  if (diP)
    INODE_UNLOCK(diP);
  if (fP)
    fput(fP);
  if (dfP)
    fput(dfP);

  TRACE4(TRACE_VNODE, 1, TRCID_KXFUNLINK_EXIT,
         "kxFlink exit: dirDesc %d fileDesc %d rc %d info: %s",
         dirDesc, fileDesc, rc, traceInfo);
  EXIT(0);
  return -rc;
}


IntRC kxFcleanWeaklink(int dirDesc,
              const char *name,
              int flags)
{
  int rc = 0;
  const char *traceInfo = "none";
  ext_cred_t  eCred;
  ext_cred_t *eCredP = &eCred;
  char nameBuffer[CXI_MAXPATHLEN + 1];
  size_t nameLength = 0;
  struct qstr nameQstr;
  Boolean caseSensitive = true;

  struct file *dfP = NULL;
  struct inode *diP = NULL;
  struct dentry *dentryP = NULL;
  struct file *fP = NULL;
  struct inode *iP = NULL;

  ENTER(0);
  TRACE3(TRACE_VNODE, 1, TRCID_KXFCLEAN_WEAKLINK_EXIT,
         "kxFcleanWeaklink enter: dirDesc %d name %lx flags 0x%X",
         dirDesc, name, flags);

  rc = getCred(&eCred, &eCredP);
  if (rc)
  {
    traceInfo = "getCred";
    goto xerror;
  }

  rc = cxiCopyInstr((char *)name, nameBuffer, sizeof(nameBuffer), &nameLength);
  if (rc != 0)
  {
    traceInfo = "cxiCopyInstr(name)";
    goto xerror;
  }

  rc = fgetValidate(dirDesc, &dfP);
  if (rc)
  {
    traceInfo = "fgetValidate(dirDesc)";
    goto xerror;
  }

  diP = dfP->f_dentry->d_inode;
  iP = fP->f_dentry->d_inode;

  nameQstr.hash = full_name_hash(nameBuffer, nameLength);
  nameQstr.len = nameLength;
  nameQstr.name = nameBuffer;
  dentryP = d_lookup(dfP->f_dentry, &nameQstr);

  rc = gpfs_ops.gpfsCleanWeaklink(VP_TO_PVP(diP), VP_TO_CNP(diP),
                                  nameBuffer, caseSensitive, eCredP);
  if (rc != 0)
  {
    traceInfo = "gpfs_ops.gpfsCleanWeaklink";
    goto xerror;
  }

xerror:
  if (dentryP)
  {
    d_delete(dentryP);
    d_drop(dentryP);
    dput(dentryP);
  }
  if (dfP)
    fput(dfP);
  putCred(&eCred, &eCredP);

    TRACE3(TRACE_VNODE, 1, TRCID_KXFCLEAN_WEAKLINK_ENTER,
         "kxFcleanWeaklink exit: dirDesc %d rc %d info: %s",
         dirDesc, rc, traceInfo);
  EXIT(0);
  return -rc;
}

#endif /* FFDEDUPE */


int kxFtruncate(int fd, offset_t length)
{
  int rc = 0;
  int code = 0;
  int flags = FWRITE;
  struct file *fP = NULL;
  ext_cred_t  eCred;
  ext_cred_t *eCredP = &eCred;
  struct MMFSVInfo *vinfoP;
  struct dentry *dentryP = NULL;
  struct dentry *parent_dentryP = NULL;
  struct inode *iP = NULL;
  struct inode *diP = NULL;
  cxiNode_t *cnFileP = NULL;
  struct gpfsVfsData_t *privVfsP;


  ENTER(0);

  TRACE2(TRACE_VNODE, 1, TRCIDFTRUNCATE1,
         "kxFtruncate enter:  fd %d, length %lld\n",
          fd, length);

  rc = getCred(&eCred, &eCredP);
  if (rc)
  {
    code = 1;
    goto xerror;
  }

  if (length < 0)
  {
    code = 2;
    rc = EINVAL;
    goto xerror;
  }

  fP = fget(fd);
  if (!fP || !fP->f_dentry || !fP->f_dentry->d_parent )
  {
    code = 3;
    rc = EBADF;
    goto xerror;
  }

  dentryP = fP->f_dentry;
  parent_dentryP = dentryP->d_parent;

  if (!dentryP->d_inode || !parent_dentryP->d_inode)
  {
    code = 4;
    rc = ENOENT;
    goto xerror;
  }
  iP = dentryP->d_inode;
  diP = parent_dentryP->d_inode;

  vinfoP = (struct MMFSVInfo *)fP->private_data;

  if (!GPFS_TYPE(iP))
  {
    code = 5;
    rc = EBADF;
    goto xerror;
  }

  privVfsP = VP_TO_PVP(iP);
  if (!privVfsP)
  {
    code = 6;
    rc = EBADF;
    goto xerror;
  }

  cnFileP = VP_TO_CNP(iP);
  if (!cnFileP)
  {
    code = 7;
    rc = EBADF;
    goto xerror;
  }

  TRACE2(TRACE_VNODE, 1, TRCIDFTRUNCATE2,
         "kxFtruncate:  iP=0x%lX, diP= 0x%lX\n",
          iP, diP);

  rc = gpfs_ops.gpfsFtrunc(privVfsP, cnFileP, flags, length, vinfoP, eCredP, true);

  if (rc != 0)
  {
    code = 8;
    goto xerror;
  }

xerror:
  putCred(&eCred, &eCredP);

  if (fP)
    fput(fP);

  TRACE3(TRACE_VNODE, 1, TRCIDFTRUNCATE3,
         "kxFtruncate exit:  fd %d, rc %d, code %d\n",
          fd, rc, code);

  EXIT(0);
  return -rc;
}

int kxSetTimes(int fd, char *pathname, int flags, void *times)
{
  int rc = 0;
  int code = 0;
  int crc;
  struct file *fP = NULL;
  struct inode *iP;
  struct nameidata nd;
  struct gpfsVfsData_t *privVfsP;
  cxiNode_t *cnP;
  ext_cred_t  eCred;
  ext_cred_t *eCredP = &eCred;
  Boolean relePath = false;

  ENTER(0);

  TRACE4(TRACE_VNODE, 1, TRCID_LINUX_SETTIMES_ENTER,
         "kxSetTimes: enter fd %d pathname 0x%x flags %d times 0x%X\n",
	 fd, pathname, flags, times);

  if (!times)
  {
    code = 2;
    rc = EINVAL;
    goto xerror;
  }

  if (pathname)
  {
    rc = USER_PATH(pathname, &nd);
    if (rc)
    {
      rc = -rc;
      code = 3;
      goto xerror;
    }
    relePath = true;

    iP = NDP_TO_IP(&nd);
    DBGASSERT(iP != NULL);
  }
  else
  {
    fP = fget(fd);
    if (!fP || !fP->f_dentry || !fP->f_dentry->d_inode)
    {
      code = 4;
      rc = EINVAL;
      goto xerror;
    }

    iP = fP->f_dentry->d_inode;
  }

  if (!iP || !(privVfsP = VP_TO_PVP(iP)))
  {
    code = 5;
    rc = EBADF;
    goto xerror;
  }

  if (!GPFS_TYPE(iP))
  {
    code = 6;
    rc = EBADF;
    goto xerror;
  }

  cnP = VP_TO_CNP(iP);

  rc = getCred(&eCred, &eCredP);
  if (rc)
    goto xerror;

  TRACE5(TRACE_VNODE, 1, TRCID_LINUX_SETTIMES_ENTER2,
         "kxSetTimes: enter fd %d fP 0x%lX inode %lld flags 0x%X arg1 0x%X\n",
	 fd, fP, iP->i_ino, flags, times);

  rc = gpfs_ops.gpfsSetTimes(flags, times, privVfsP, cnP, eCredP);

xerror:
  putCred(&eCred, &eCredP);

  if (relePath)
    PATH_PUT(&nd);

  if (fP)
    fput(fP);

  TRACE2(TRACE_VNODE, 1, TRCID_LINUX_SETTIMES_EXIT,
         "kxSetTimes: exit rc %d code %d\n", rc, code);

  EXIT(0);
  return -rc;
}


#ifdef GANESHA

#if (LINUX_KERNEL_VERSION >= 2062400)
struct dentry *
gpfs_fh_to_dentry(struct super_block *sbP, struct fid *fid,
                  int len, int fhtype);
#endif

#if (LINUX_KERNEL_VERSION >= 2062500)
#define USER_WALK_PATH user_path_at
#define ND_DENTRY(nd) nd.path.dentry
#define PWDMNT pwd.mnt
#define FSUID current_fsuid()
#define FGET(fd, fput_needed) fget(fd)
#define FPUT(fp, fput_needed) fput(fp)
#else
#define USER_WALK_PATH __user_walk_fd
#define ND_DENTRY(nd) nd.dentry
#define PWDMNT pwdmnt
#define FSUID current->fsuid
#if defined(REDHAT_AS_LINUX)
#define FGET(fd, fput_needed) fget_light(fd, &fput_needed)
#define FPUT(fp, fput_needed) fput_light(fp, fput_needed)
#else
#define FGET(fd, fput_needed) fget(fd)
#define FPUT(fp, fput_needed) fput(fp)
#endif
#endif

#define FH_HDR_SIZE (sizeof(struct gpfs_file_handle)-OPENHANDLE_HANDLE_LEN)

struct handle_svc_export
{
  struct cache_head h;
  struct auth_domain *ex_client;
  int ex_flags;
};

typedef union {
  struct open_arg         oarg;
  struct open_share_arg   osarg;
  struct share_reserve_arg sarg;
  struct link_arg         linkarg;
  struct stat_name_arg    snarg;
  struct link_fh_arg      lnarg;
  struct rename_fh_arg    rnarg;
  struct name_handle_arg  harg;
  struct get_handle_arg   gharg;
  struct readlink_arg     readlinkarg;
  struct readlink_fh_arg  rlfharg;
  struct stat_arg         statarg;
  struct deviceinfo_arg   darg;
  struct layoutget_arg    yarg;
  struct layoutreturn_arg zarg;
  struct dsread_arg       rarg;
  struct dswrite_arg      warg;
  struct xstat_access_arg xarg;
  struct xstat_arg        carg;
  struct callback_arg     barg;
  struct set_get_lock_arg larg;
  struct layoutcommit_arg targ;
  struct fsync_arg        farg;
  struct close_file_arg   cfarg;
} all_args;

#define MY_ACC_MODE(x) ("\000\004\002\006"[(x)&O_ACCMODE])

static int vfs_dentry_acceptable(void *context, struct dentry *dentry)
{
  return 1;
}

int posix2glock(struct glock *glock, struct file_lock *plock);

static int ganesha_grant_deferred(struct file_lock *fl, struct file_lock *conf,
                                  int result)
{
  int rc = -ENOENT;
  int reason;
  struct glock glock;
  struct flock *flock = &glock.flock;
  cxiNode_t *cnP;
  struct gpfsVfsData_t *privVfsP;
  struct file *fP;

  TRACE3(TRACE_VNODE, 3, TRCID_GANESHA_GRANT_LOCK,
         "ganesha_grant_deferred: fl 0x%lX conf 0x%lX result %d",
          fl, conf, result);

  if (result == EAGAIN)
    reason = INODE_LOCK_AGAIN;
  else
    reason = INODE_LOCK_GRANTED;

  if (!result || result == EAGAIN)
  {
    fP = fl->fl_file;
    DBGASSERT(fP != NULL);
    cnP = VP_TO_CNP(fP->f_dentry->d_inode);
    privVfsP = VP_TO_PVP(fP->f_dentry->d_inode);
    DBGASSERT(privVfsP != NULL);
    rc = posix2glock(&glock, fl);

    TRACE6(TRACE_VNODE, 3, TRCID_LINUX_GRANT_LOCK_06,
        "ganesha_grant_deferred: rc %d lock pid %d type %d start %lld len %lld lock_owner 0x%lX",
         rc, flock->l_pid, flock->l_type, flock->l_start,
         flock->l_len, glock.lock_owner);

    if (!rc)
      rc = gpfs_ops.gpfsGaneshaLock(privVfsP, cnP, &glock, fl, reason);

    if (rc)
      LOCK_FREE_LOCK(fl);
  }
  return rc;
}

static void ganesha_break_deleg(struct file_lock *fl)
{
  int rc = -ENOENT;
  int reason;
  struct glock glock;
  struct flock *flock = &glock.flock;
  cxiNode_t *cnP;
  struct gpfsVfsData_t *privVfsP;
  struct file *fP;

  reason = BREAK_DELEGATION;

  fP = fl->fl_file;
  DBGASSERT(fP != NULL);
  cnP = VP_TO_CNP(fP->f_dentry->d_inode);
  privVfsP = VP_TO_PVP(fP->f_dentry->d_inode);
  DBGASSERT(privVfsP != NULL);
  rc = posix2glock(&glock, fl);

  TRACE6(TRACE_VNODE, 3, TRCID_LINUX_BREAK_LOCK_06,
      "ganesha_break_deleg: rc %d lock pid %d type %d start %lld len %lld lock_owner 0x%lX",
       rc, flock->l_pid, flock->l_type, flock->l_start,
       flock->l_len, glock.lock_owner);

  if (!rc)
    rc = gpfs_ops.gpfsGaneshaLock(privVfsP, cnP, &glock, fl, reason);

  return;
}

static int ganesha_change_deleg(struct file_lock **fl, int arg)
{
  int rc = 0;

  if (arg & F_UNLCK)
    rc = lease_modify(fl, arg);
  else
    rc = -EAGAIN;

  TRACE3(TRACE_VNODE, 3, TRCID_GANESHA_CHANGE_LOCK,
         "ganesha_change_deleg: fl 0x%lX arg %d rc %d",
          fl, arg, rc);

  return rc;
}

static void ganesha_notify_deferred(struct file_lock *fl)
{
  struct file *filp = NULL;
  int rc;

  TRACE2(TRACE_VNODE, 3, TRCID_GANESHA_NOTIFY_LOCK,
         "ganesha_notify_deferred: fl 0x%lX filp 0x%lX", fl, filp);

  rc = ganesha_grant_deferred(fl, NULL, EAGAIN);

  TRACE3(TRACE_VNODE, 3, TRCID_GANESHA_NOTIFY_LOCK_2,
         "ganesha_notify_deferred: rc %d fl 0x%lX filp 0x%lX", rc, fl, filp);

  return;
}

static int ganesha_same_owner(struct file_lock *fl1, struct file_lock *fl2)
{
  TRACE6(TRACE_VNODE, 5, TRCID_LINUX_SAMET_LOCK_06,
         "ganesha_same_owner flags %x:%x pid %d:%d owner 0x%lX:0x%lX",
          fl1->fl_flags, fl2->fl_flags, fl1->fl_pid, fl2->fl_pid,
          fl1->fl_owner, fl2->fl_owner);

#ifdef HAS_FL_CLOSE
  if (fl1->fl_flags & FL_CLOSE)
  {
    /* If this is a close for Ganesha that we want to keep the locks for fl_owner is NULL */
    if (fl1->fl_owner == NULL)
      return 0;
    return fl1->fl_pid == fl2->fl_pid;
  }
  else
#endif
    return fl1->fl_owner == fl2->fl_owner && fl1->fl_pid == fl2->fl_pid;
}

struct lock_manager_operations ganesha_lock_operations = {
	.fl_compare_owner = ganesha_same_owner,
	.fl_notify = ganesha_notify_deferred,
#ifdef NFS_CLUSTER_LOCKS
	.fl_grant = ganesha_grant_deferred,
        .fl_break = ganesha_break_deleg,
        .fl_change = ganesha_change_deleg,
#if LINUX_KERNEL_VERSION < 2063800
	.fl_mylease = ganesha_same_owner,
#endif
#endif
};

int ganesha_tgid = 0;

static int check_type(struct file_lock *flock, int type)
{
  switch (type)
  {
    case F_WRLCK:
    case F_RDLCK:
    case F_UNLCK:
      flock->fl_type = type;
      break;

    default:
      return -EINVAL;
  }
  return 0;
}

static int glock2posix(struct file *fP, struct file_lock *plock, struct glock *g)
{
  off_t first, last;
  int retval = 0;
  int code = 0;

  switch (g->flock.l_whence)
  {
    case SEEK_SET:
      first = 0;
      break;
    case SEEK_CUR:
      first = fP->f_pos;
      break;
    default:
     {
       code = 1;
       retval = -EINVAL;
       goto out;
     }
  }
  first += g->flock.l_start;
  if (first < 0)
  {
    code = 3;
    retval = -EINVAL;
    goto out;
  }
  plock->fl_end = OFFSET_MAX;
  if (g->flock.l_len > 0)
  {
    last = first + g->flock.l_len - 1;
    plock->fl_end = last;
  }
  else if (g->flock.l_len < 0)
  {
    last = first - 1;
    plock->fl_end = last;
    first += g->flock.l_len;
    if (first < 0)
    {
      code = 5;
      retval = -EINVAL;
      goto out;
    }
  }
  plock->fl_start = first;	
  if (plock->fl_end < plock->fl_start)
  {
    code = 7;
    retval = -EOVERFLOW;
    goto out;
  }
  if (g->lock_owner != NULL)
    plock->fl_owner = g->lock_owner+1;  // make sure it is not a valid address.
  else
    plock->fl_owner = g->lock_owner;

  if (ganesha_tgid == 0)
    ganesha_tgid = cxiGetProcessId();
  else
  {
    if (ganesha_tgid != cxiGetProcessId())
    {
      code = 9;
      retval = -EINVAL;
      goto out;
    }
  }
  plock->fl_pid = cxiGetProcessId();
  plock->fl_file = fP;
  plock->fl_flags = FL_POSIX;
  plock->fl_ops = NULL;
  plock->fl_lmops = &ganesha_lock_operations;

  retval = check_type(plock, g->flock.l_type);

out:
  TRACE2(TRACE_VNODE, 1, TRCID_LINUX_GLOCK_TO_POSIX,
         "glock2posix: retval %d code %d", retval, code);

  return retval;
}

int posix2glock(struct glock *glock, struct file_lock *plock)
{
#if BITS_PER_LONG == 32
  if (plock->fl_start > OFFT_OFFSET_MAX)
    return -EOVERFLOW;
  if (plock->fl_end != OFFSET_MAX && plock->fl_end > OFFT_OFFSET_MAX)
    return -EOVERFLOW;
#endif
  glock->flock.l_start = plock->fl_start;
  glock->flock.l_len = plock->fl_end == OFFSET_MAX?0:plock->fl_end-plock->fl_start + 1;
  glock->flock.l_whence = 0;
  glock->flock.l_pid = plock->fl_pid;
  glock->flock.l_type = plock->fl_type;
  glock->lock_owner = (void *)(((long)plock->fl_owner >> 1) << 1); // make even
  return 0;
}

int
gpfs_wait_inode_update(struct super_block *sbP,
                       struct gpfs_file_handle __user *ufh, void __user *statbuf,
                       struct glock __user *glock, int *reasonP, int *flagsP)
{
  struct gpfsVfsData_t *privVfsP;
  struct gpfs_file_handle fh;
  int retval = 0;
  int len = 0;
  int reason = 0;
  int code = 0;
  unsigned int *fhP;
  cxiVattr_t vattr;
  struct glock tglock;
#if BITS_PER_LONG != 64
  struct stat64 tmp;
#else
  struct stat tmp;
#endif

  if (copy_from_user(&fh, ufh, FH_HDR_SIZE))
  {
    code = 1;
    retval = EFAULT;
    goto err_out;
  }
  len = fh.handle_size;

  ganesha_tgid = cxiGetProcessId();

  TRACE4(TRACE_VNODE, 1, TRCID_LINUX_GABESHA_INODE_01,
         "gpfs_wait_inode_update: ganesha_tgid %d gid %d len %d sbP 0x%lX",
          ganesha_tgid, GET_GROUP_ID(current), len, sbP);

  privVfsP = (struct gpfsVfsData_t *)cxiGetPrivVfsP(sbP);

retry:
  retval = gpfs_ops.gpfsGaneshaUpdate(privVfsP, (int *)&fh.f_handle, &len,
                                      &vattr, &tglock, &reason);
  if (retval)
  {
    if (retval == EAGAIN)
      goto retry;
    code = 3;
    goto err_out;
  }
  fh.handle_size = len;
  if (len == (MAX_FH_INT >> 2))
    fh.handle_key_size = 24;  // 40 - 16 don't include parent info
  else
    fh.handle_key_size = 12;  // 32 bit inode

  if (copy_to_user(ufh, &fh, FH_HDR_SIZE+len))
  {
    code = 5;
    retval = EFAULT;
    goto err_out;
  }
  if (copy_to_user(reasonP, &reason, sizeof(int)))
  {
    code = 7;
    retval = EFAULT;
    goto err_out;
  }
  fhP = (int *)&fh.f_handle;

  TRACE8(TRACE_VNODE, 3, TRCID_ENCODE_FH_1,
         "gpfs_wait_inode_update len %d: %08x %08x %08x %08x %08x %08x %08x",
          len, fhP[0],fhP[1],fhP[2],fhP[3],fhP[4],fhP[5],fhP[6]);

  tmp.st_ino = vattr.va_ino;

  if (reason == INODE_INVALIDATE)
  {
    code = 11;
  }
  if (reason == INODE_UPDATE)
  {
    if (copy_to_user(flagsP, &vattr.va_flags, sizeof(int)))
    {
      code = 12;
      retval = EFAULT;
      goto err_out;
    }
    tmp.st_nlink = (Int64)vattr.va_nlink;
    tmp.st_atime = vattr.va_atime.tv_sec;
    tmp.st_mtime = vattr.va_mtime.tv_sec;
    tmp.st_ctime = vattr.va_ctime.tv_sec;
#ifdef STAT_HAVE_NSEC
    tmp.st_atime_nsec = vattr.va_atime.tv_nsec;
    tmp.st_mtime_nsec = vattr.va_mtime.tv_nsec;
    tmp.st_ctime_nsec = vattr.va_ctime.tv_nsec;
#endif
    code = 13;
  }
  if (reason == INODE_LOCK_GRANTED || reason == INODE_LOCK_AGAIN)
  {
     TRACE5(TRACE_VNODE, 3, TRCID_LINUX_GABESHA_INODE_06,
          "gpfs_wait_inode_update: lock pid %d type %d start %lld len %lld lock_owner0x%lX",
           tglock.flock.l_pid, tglock.flock.l_type, tglock.flock.l_start,
           tglock.flock.l_len, tglock.lock_owner);
    if (copy_to_user(glock, &tglock, sizeof(struct glock)))
    {
      code = 17;
      retval = EFAULT;
      goto err_out;
    }
    if (reason == INODE_LOCK_AGAIN)
    {
      code = 19;
      retval = ENOLCK;
      goto err_out;
    }
  }
  if (copy_to_user(statbuf, &tmp, sizeof(tmp)))
  {
    code = 21;
    retval = EFAULT;
    goto err_out;
  }

err_out:
  TRACE7(TRACE_VNODE, 1, TRCID_LINUX_GABESHA_INODE_04,
     "gpfs_wait_inode_update:  retval %d reason %d ino %lld nlink %lld code %d len %d key size %d",
      retval, reason, tmp.st_ino, tmp.st_nlink, code, len, fh.handle_key_size);

  return (-retval);
}

int
gpfs_wait_update(struct super_block *sbP, int *reasonP)
{
  struct gpfsVfsData_t *privVfsP;
  int retval = 0;
  int reason = 0;
  int code = 0;

  ganesha_tgid = cxiGetProcessId();

  if (copy_from_user(&reason, reasonP, sizeof(int)))
  {
    retval = -EFAULT;
    goto err_out;
  }

  TRACE4(TRACE_VNODE, 1, TRCID_LINUX_GABESHA_UPDATE_01,
         "gpfs_wait_update: ganesha_tgid %d gid %d sbP 0x%lX reason %d",
          ganesha_tgid, GET_GROUP_ID(current), sbP, reason);

  privVfsP = (struct gpfsVfsData_t *)cxiGetPrivVfsP(sbP);

  retval = gpfs_ops.gpfsGaneshaThread(privVfsP, &reason);
  if (retval)
  {
    code = 3;
    goto err_out;
  }
  if (copy_to_user(reasonP, &reason, sizeof(int)))
  {
    code = 7;
    retval = EFAULT;
    goto err_out;
  }

err_out:
  TRACE3(TRACE_VNODE, 1, TRCID_LINUX_GABESHA_UPDATE_03,
     "gpfs_wait_update:  retval %d reason %d code %d",
      retval, reason, code);

  return (-retval);
}

static int dentry_to_handle(struct dentry *dentry,
                            struct gpfs_file_handle __user *ufh)
{
  int retval;
  int handle_size = 0;
  struct gpfs_file_handle fh;
  const struct export_operations *exop = dentry->d_sb->s_export_op;

  if (copy_from_user(&fh, ufh, FH_HDR_SIZE))
  {
    retval = -EFAULT;
    goto err_out;
  }
  if (fh.handle_size > OPENHANDLE_HANDLE_LEN)
  {
    retval = -EINVAL;
    goto err_out;
  }
  handle_size = fh.handle_size >> 2;
  /* we ask for a non connected handle */
  retval = gpfs_encode_fh(dentry, (int *)&fh.f_handle, &handle_size, 0);
  /* convert handle size to bytes */
  handle_size = handle_size << 2;
  fh.handle_size = handle_size;
  if (handle_size == (MAX_FH_INT >> 2))
    fh.handle_key_size = 24;  // 40 - 16 don't include parent info
  else
    fh.handle_key_size = 12;  // 32 bit inode

  if (retval == 255)
    retval = -ENOSPC;
  else
  {
    fh.handle_type = retval;
    retval = 0;
    if (copy_to_user(ufh, &fh, FH_HDR_SIZE+handle_size))
    {
      retval = -EFAULT;
      goto out;
    }
  }
  TRACE3(TRACE_VNODE, 3, TRCID_GANESHA_NAME_TO_HANDLE_OUT,
         "dentry_to_handle: exit retval %d handle_size %d key_size %d",
          retval, fh.handle_size, fh.handle_key_size);

  goto out;

err_out:
  TRACE2(TRACE_VNODE, 3, TRCID_GANESHA_NAME_TO_HANDLE_EXIT,
         "dentry_to_handle: exit retval %d handle_size %d",
          retval, fh.handle_size);
out:
  return retval;
}

int name_to_handle_at(int dfd, const char __user *name,
                       struct gpfs_file_handle __user *handle, int flag)
{
  int follow;
  int ret = -EINVAL;
  struct nameidata nd;
  struct file *file = NULL;

  if (!capable(CAP_DAC_OVERRIDE))
  {
    ret = -EPERM;
    goto err_out;
  }
#if LINUX_KERNEL_VERSION < 2061800 || defined(SUSE_SLES10SP2) || defined(SUSE_SLES11) || defined(UBUNTU_LINUX)
  ret = -ENOENT; //??? fix for old SLES
  goto err_out;
#else
  if ((flag & ~AT_SYMLINK_FOLLOW) != 0)
    goto err_out;
#endif

  if (name == NULL && dfd != AT_FDCWD)
  {
    file = fget(dfd);

    if (file)
      ND_PATH(nd).dentry = file->f_dentry;
    else
    {
      ret = -EBADF;
      goto err_out;
    }
  }
  else
  {
#if LINUX_KERNEL_VERSION < 2061800 || defined(SUSE_SLES10SP2) || defined(SUSE_SLES11) || defined(UBUNTU_LINUX)
    ret = -ENOENT; //??? fix for old SLES
    goto err_out;
#else
    follow = (flag & AT_SYMLINK_FOLLOW) ? LOOKUP_FOLLOW : 0;
#endif
    ret = USER_WALK_PATH(dfd, name, follow, &ND_PATH(nd));
    if (ret)
      goto err_out;
  }
  ret = dentry_to_handle(ND_DENTRY(nd), handle);
  if (file)
    fput(file);
  else
    PATH_PUT(&nd);

err_out:

  TRACE2(TRACE_VNODE, 3, TRCID_GANESHA_HANDLE_AT_EXIT,
         "name_to_handle_at: exit ret %d dfd %d",
          ret, dfd);

  return ret;
}

#if (LINUX_KERNEL_VERSION >= 2063200)
static struct path *get_path_from_fd(int fd)
{
  int fput_needed = 0;
  struct path *path;
  struct file *filep;

  if (fd == AT_FDCWD)
  {
    struct fs_struct *fs = current->fs;
    /* spin_lock or read_lock depends on LINUX_KERNEL_VERSION */
    FS_STRUCT_LOCK(&fs->lock);
    path = &fs->pwd;
    path_get(path);
    FS_STRUCT_UNLOCK(&fs->lock);
  }
  else
  {
    filep = FGET(fd, fput_needed);
    if (!filep)
      return ERR_PTR(-EBADF);
    path = &filep->f_path;
    path_get(path);
    FPUT(filep, fput_needed);
  }
  return path;
}

static struct vfsmount *get_vfsmount_from_fd(int fd)
{
  struct path *path;
  struct vfsmount *mnt = ERR_PTR(-EBADF);

  path = get_path_from_fd(fd);

  if (!IS_ERR(path))
  {
    mnt = path->mnt;
    mntget(mnt);
    path_put(path);
  }
  return mnt;
}

#else

static struct vfsmount *get_vfsmount_from_fd(int fd)
{
  int fput_needed;
  struct vfsmount *mnt;
  struct file *filep;

  if (fd == AT_FDCWD)
  {
    struct fs_struct *fs = current->fs;
    read_lock(&fs->lock);
    mnt = fs->PWDMNT;
    mntget(mnt);
    read_unlock(&fs->lock);
  }
  else
  {
    filep = FGET(fd, fput_needed);
    if (!filep)
      return ERR_PTR(-EBADF);
    mnt = filep->f_vfsmnt;
    mntget(mnt);
    FPUT(filep, fput_needed);
  }
  return mnt;
}
#endif

static struct dentry *
getDentryFromFH(struct super_block *sbP, struct gpfs_file_handle *handle)
{
  int handle_size;
  struct dentry *dentry;
  struct handle_svc_export svc;

  /* change the handle size to multiple of sizeof(u32) */
  handle_size = handle->handle_size >> 2;

#if (LINUX_KERNEL_VERSION >= 2062400)
  dentry = gpfs_fh_to_dentry(sbP, (struct fid *) handle->f_handle,
                             handle_size, handle->handle_type);
#else
  /*
   * GPFS overload acceptable callback. To make GPFS return
   * 1 set svc->ex_flags
   */
  memset(&svc, 0, sizeof(svc));
  svc.ex_flags = NFSEXP_NOSUBTREECHECK;

  dentry = gpfs_decode_fh(sbP, (__u32 *) handle->f_handle,
                          handle_size, handle->handle_type,
                          vfs_dentry_acceptable, (void *)&svc);
#endif
  return dentry;
}

static struct dentry *handle_to_dentry(int mountdirfd,
                         struct gpfs_file_handle *handle, struct vfsmount **mntp)
{
  long retval;
  int code = 0;
  struct vfsmount *mnt;
  struct dentry *dentry;
  struct handle_svc_export svc;

  mnt = get_vfsmount_from_fd(mountdirfd);
  if (IS_ERR(mnt))
  {
    code = 2;
    retval = PTR_ERR(mnt);
    goto out_err;
  }
  dentry = getDentryFromFH(mnt->mnt_sb, handle);
  if (IS_ERR(dentry))
  {
    code = 4;
    retval = PTR_ERR(dentry);
    goto out_mnt;
  }
  if (!dentry)
  {
    code = 6;
    retval = -ESTALE;
    goto out_mnt;
  }
  *mntp = mnt;
  return dentry;

out_mnt:
  mntput(mnt);
out_err:

  TRACE3(TRACE_VNODE, 3, TRCID_GANESHA_HANDLE_TO_DENTRY_EXIT,
         "handle_to_dentry: exit err %ld retval %lx code %d",
          ERR_PTR(retval), retval, code);

  return ERR_PTR(retval);
}

int suidRemove(struct dentry *dentry)
{
  mode_t i_mode = dentry->d_inode->i_mode;
  int kill_attr = 0;

  if (i_mode & S_ISUID)
    kill_attr = ATTR_KILL_SUID;

  if ((i_mode & S_ISGID) && (i_mode & S_IXGRP))
    kill_attr |= ATTR_KILL_SGID;

  if (kill_attr && !capable(CAP_FSETID))
    return kill_attr;

  return 0;
}

int truncate_it(struct dentry *dentry, loff_t len, unsigned int attr_time)
{
  int rc;
  struct iattr attrs;

  if (len < 0)
    return -EINVAL;

  attrs.ia_size = len;
  attrs.ia_valid = ATTR_SIZE | attr_time;

  attrs.ia_valid |= suidRemove(dentry);

  mutex_lock(&dentry->d_inode->i_mutex);
#if defined(SUSE_SLES10SP2) || defined(SUSE_SLES11) || defined(UBUNTU_LINUX)
  rc = -ENOENT; //??? fix for old SLES
#else
  rc = notify_change(dentry, &attrs);
#endif
  mutex_unlock(&dentry->d_inode->i_mutex);

  TRACE3(TRACE_VNODE, 3, TRCID_GANESHA_HANDLE_TRUNCATE_EXIT,
         "truncate_it: exit rc %d inode %lld len %lld",
          rc, dentry->d_inode->i_ino, len);

  return rc;
}

int check_permission(struct inode *inode, int mask, unsigned int flags)
{
  int submask;
  umode_t i_mode = inode->i_mode;

  if (mask & MAY_WRITE)
  {
    if (IS_RDONLY(inode) && (S_ISREG(i_mode) || S_ISDIR(i_mode) || S_ISLNK(i_mode)))
      return -EROFS;

    if (IS_IMMUTABLE(inode))
      return -EACCES;
  }

  if ((mask & MAY_EXEC) && S_ISREG(i_mode) && !(i_mode & S_IXUGO))
    return -EACCES;

  submask = mask & ~MAY_APPEND;
#ifdef IN_KERNEL_CHANGE_NOT_SUPP

  /* FIXME! we don't have nameidata for handle lookup. So not sure how
   * we can check for inode->permissions. We already limit the call
   * to CAP_DAC_OVERRIDE. So we should be able to skip the ACL check.
   * But we also skip security_inode_permission below.
   */

  if (inode->i_op && inode->i_op->permission)
    retval = inode->i_op->permission(inode, submask, nd);
  else
    retval = LINUX_GENERIC_PERMISSION(inode, submask, flags, NULL);
  if (retval)
    return retval;

  return security_inode_permission(inode, mask, nd);
#else
  return LINUX_GENERIC_PERMISSION(inode, submask, flags, NULL);
#endif
}

static int may_handle_open(struct dentry *dentry, int open_flag)
{
  int acc_mode;
  int error;
  int code = 0;
  struct inode *inode = dentry->d_inode;

  if ((open_flag + 1) & O_ACCMODE)
    open_flag++;

  acc_mode = MY_ACC_MODE(open_flag);

  if (open_flag & O_TRUNC)
    acc_mode |= MAY_WRITE;

  if (open_flag & O_APPEND)
    acc_mode |= MAY_APPEND;

  TRACE3(TRACE_VNODE, 3, TRCID_GANESHA_MAY_HANDLE_OPEN,
         "may_handle_open: enter inode %lld acc_mode %X open_flag %X",
          inode->i_ino, acc_mode, open_flag);

  if (S_ISDIR(inode->i_mode) && (acc_mode & MAY_WRITE))
  {
    error = -EISDIR;
    code = 3;
    goto out;
  }
  error = check_permission(inode, acc_mode, open_flag);
  if (error)
  {
    code = 5;
    goto out;
  }

  if (S_ISFIFO(inode->i_mode) || S_ISSOCK(inode->i_mode))
  {
    open_flag &= ~O_TRUNC;
  }
  else if (S_ISBLK(inode->i_mode) || S_ISCHR(inode->i_mode))
  {
#ifdef IN_KERNEL_CHANGE_NOT_SUPP
    if (nd->mnt->mnt_flags & MNT_NODEV)
    {
      error = -EACCES;
      code = 7;
      goto out;
    }
#endif
    open_flag &= ~O_TRUNC;
  }
  else if (IS_RDONLY(inode) && (acc_mode & MAY_WRITE))
  {
    error = -EROFS;
    code = 9;
    goto out;
  }
  if (IS_APPEND(inode))
  {
    if ((open_flag & FMODE_WRITE) && !(open_flag & O_APPEND))
    {
      error = -EPERM;
      code = 11;
      goto out;
    }
    if (open_flag & O_TRUNC)
    {
      error = -EPERM;
      code = 13;
      goto out;
    }
  }
  if (open_flag & O_NOATIME)
    if (FSUID != inode->i_uid && !capable(CAP_FOWNER))
    {
      error = -EPERM;
      code = 15;
      goto out;
    }

  error = break_lease(inode, open_flag);
  if (error)
  {
    code = 17;
    goto out;
  }
  if (open_flag & O_TRUNC)
  {
    error = get_write_access(inode);
    if (error)
    {
      code = 19;
      goto out;
    }

#ifdef IN_KERNEL_CHANGE_NOT_SUPP
    error = locks_verify_locked(inode);
#endif
    if (!error)
    {
      error = truncate_it(dentry, 0, ATTR_MTIME | ATTR_CTIME);
    }
    put_write_access(inode);
    if (error)
    {
      code = 21;
      goto out;
    }
  }
  error = 0;

out:
  TRACE3(TRACE_VNODE, 3, TRCID_GANESHA_MAY_HANDLE_OPEN_EXIT,
         "may_handle_open: exit inode %lld error %d code %d",
          inode->i_ino, error, code);

  return error;
}

static struct dentry *
dentry_by_handle(int mountdirfd, struct gpfs_file_handle __user *ufh)
{
  int retval = 0;
  struct dentry *dentry = NULL;
  struct vfsmount *mnt = NULL;
  struct gpfs_file_handle fh;

  if (copy_from_user(&fh, ufh, FH_HDR_SIZE))
    return ERR_PTR(-EFAULT);

  if ((fh.handle_size > OPENHANDLE_HANDLE_LEN) || (fh.handle_size <= 0))
    return ERR_PTR(-EINVAL);

  /* copy the full handle */
  if (copy_from_user(&fh, ufh, FH_HDR_SIZE + fh.handle_size))
    return ERR_PTR(-EFAULT);

  dentry = handle_to_dentry(mountdirfd, &fh, &mnt);
  if (IS_ERR(dentry))
    return dentry;

  mntput(mnt);

  return dentry;
}

int do_sys_open_by_handle(int mountdirfd, struct gpfs_file_handle __user *ufh,
                          int open_flag, int share_access, int share_deny)
{
  int fd = 0;
  int retval = 0;
  struct file *filp;
  struct dentry *dentry;
  struct vfsmount *mnt = NULL;
  struct gpfs_file_handle fh;

  /* can't use O_CREATE with open_by_handle */
  if (open_flag & O_CREAT)
  {
    retval = -EINVAL;
    goto out;
  }
  if (copy_from_user(&fh, ufh, FH_HDR_SIZE))
  {
    retval = -EFAULT;
    goto out;
  }
  if ((fh.handle_size > OPENHANDLE_HANDLE_LEN) || (fh.handle_size <= 0))
  {
    retval = -EINVAL;
    goto out;
  }
  /* copy the full handle */
  if (copy_from_user(&fh, ufh, FH_HDR_SIZE + fh.handle_size))
  {
    retval = -EFAULT;
    goto out;
  }
  dentry = handle_to_dentry(mountdirfd, &fh, &mnt);
  if (IS_ERR(dentry))
  {
    retval = PTR_ERR(dentry);
    goto out;
  }
  retval = may_handle_open(dentry, open_flag);
  if (retval)
    goto out_dentry;

  fd = get_unused_fd();
  if (fd < 0)
  {
    retval = fd;
    goto out_dentry;
  }
  filp = DENTRY_OPEN(dentry, mnt, open_flag);
  if (IS_ERR(filp))
  {
    put_unused_fd(fd);
    retval = PTR_ERR(filp);
  }
  else
  {
    retval = fd;
    FSNOTIFY_OPEN(filp);
    fd_install(fd, filp);
  }
  if (retval > 0)
  {
    int err = 0;

    if ((share_access == 0) && (share_deny == 0))  /* no shares */
      goto out;

    err = gpfs_f_share(filp, share_access, share_deny);

    TRACE4(TRACE_VNODE, 3, TRCID_GANESHA_DO_SYS_OPEN_SHARE,
         "do_sys_open_by_handle: fd %d err %d share_access %d share_deny %d",
          fd, err, share_access, share_deny);

    if (err)
    {
      fput(filp);
      retval = -(err);
      goto out_dentry;
    }
  }
  goto out;

out_dentry:
  dput(dentry);
  mntput(mnt);

out:
  TRACE3(TRACE_VNODE, 3, TRCID_GANESHA_DO_SYS_OPEN_EXIT,
         "do_sys_open_by_handle: exit retval %d mountdirfd %d fd %d",
          retval, mountdirfd, fd);

  return retval;
}

int open_by_handle(int mountdirfd, struct gpfs_file_handle __user *handle,
                   int flags, int share_access, int share_deny)
{
  int ret;

  if (!capable(CAP_DAC_OVERRIDE))
    return -EPERM;

  if (force_o_largefile())
    flags |= O_LARGEFILE;

  ret = do_sys_open_by_handle(mountdirfd, handle, flags,
                              share_access, share_deny);
  return ret;
}

/* get/set lock or delegation */

int set_lock(int mountdirfd, struct glock __user *l, Boolean set_delegation)
{
  struct file_lock *flock = NULL;
  struct file_lock *old_fl = NULL;
  struct glock glock;
  struct file *filep = NULL;
  struct file *f;
  struct inode *iP = NULL;
  int fput_needed = 0;
  int error = 0;
  int code = 0;

  TRACE2(TRACE_VNODE, 3, TRCID_LINUXOPS_SET_DELEG_IN,
        "set_lock: dirfd %d set_delegation %d",
         mountdirfd, set_delegation);

#ifndef NFS_CLUSTER_LOCKS
   code = 2;
   error = -ENOLCK;
   goto out;
#endif
  if (!capable(CAP_DAC_OVERRIDE))
  {
    code = 1;
    error = -EPERM;
    goto out;
  }
  if (copy_from_user(&glock, l, sizeof(glock)))
  {
    code = 3;
    error = -EFAULT;
    goto out;
  }
  filep = FGET(glock.lfd, fput_needed);
  if (!filep)
  {
    code = 5;
    error = -EBADF;
    goto out;
  }
  flock = LOCK_ALLOC_LOCK();
  if (flock == NULL)
  {
    code = 7;
    error = -ENOMEM;
    goto out_fd;
  }
  old_fl = flock;

  TRACE6(TRACE_VNODE, 3, TRCID_SET_LOCK_1,
         "set_lock: fl 0x%lX dfd %d lfd %d cmd %d owner 0x%lX type %d",
          flock, mountdirfd, glock.lfd, glock.cmd, glock.lock_owner,
          glock.flock.l_type);

  locks_init_lock(flock);
  error = glock2posix(filep, flock, &glock);
  if (error)
  {
    code = 9;
    error = -EBADF;
    goto out_fd;
  }
  if (set_delegation)
  {
    flock->fl_flags = FL_LEASE;
    if (flock->fl_type != F_RDLCK && flock->fl_type != F_UNLCK)
    {
      code = 11;
      error = -ENOLCK;
      goto out_fd;
    }
#ifdef NFS_CLUSTER_LOCKS
    error = gpfs_f_set_lease(filep, flock->fl_type, &flock);
#endif
    if (error == 0)   /* 0 means we got the delegation, keep flock */
    {
      /* If the flock returned is the same as the one we allocated we don't free it */
      if (flock->fl_type != F_UNLCK && flock == old_fl)
        old_fl = NULL;
      code = 21;
      goto out_fd;

    }
    else
    {
      code = 23;
      goto out_fd;
    }
  }
  else
  {
    if (glock.cmd == F_SETLKW)
      flock->fl_flags |= FL_SLEEP;
#ifdef NFS_CLUSTER_LOCKS
    if (glock.cmd == GPFS_F_CANCELLK)
      glock.cmd == F_CANCELLK;
#endif

    error = -EBADF;
    switch (glock.flock.l_type)
    {
      case F_RDLCK:
        if (!(filep->f_mode & FMODE_READ))
          goto out_fd;
        break;
      case F_WRLCK:
        if (!(filep->f_mode & FMODE_WRITE))
          goto out_fd;
        break;
      case F_UNLCK:
        break;
      default:
        error = -EINVAL;
        goto out_fd;
    }
    error = gpfs_f_lock(filep, glock.cmd, flock);
    if (error == 1)   /* 1 means we will get callback, keep flock */
      old_fl = NULL;
  }
out_fd:

  if (old_fl)
    LOCK_FREE_LOCK(old_fl);

  FPUT(filep, fput_needed);

out:

  TRACE7(TRACE_VNODE, 3, TRCID_SET_LOCK_EXIT_1,
   "set_lock: exit dfd %d lfd %d cmd %d old_fl 0x%lX fl 0x%lX error %d code %d",
    mountdirfd, glock.lfd, glock.cmd, old_fl, flock, error, code);

  return error;
}

int get_lock(int mountdirfd, struct glock __user *l)
{
  struct file_lock file_lock;
  struct glock glock;
  struct file *filep = NULL;
  int fput_needed = 0;
  int error = 0;
  int code = 0;

  if (!capable(CAP_DAC_OVERRIDE))
  {
    code = 1;
    error = -EPERM;
    goto out;
  }
  if (copy_from_user(&glock, l, sizeof(glock)))
  {
    code = 3;
    error = -EFAULT;
    goto out;
  }
  TRACE4(TRACE_VNODE, 3, TRCID_GET_LOCK_1,
         "get_lock: dfd %d lfd %d cmd %d lock_owner 0x%lX",
          mountdirfd, glock.lfd, glock.cmd, glock.lock_owner);

  filep = FGET(glock.lfd, fput_needed);
  if (!filep)
  {
    code = 5;
    error = -EBADF;
    goto out;
  }
  if ((glock.flock.l_type != F_RDLCK) && (glock.flock.l_type != F_WRLCK))
  {
    code = 7;
    error = -EINVAL;
    goto out_fd;
  }
  locks_init_lock(&file_lock);
  error = glock2posix(filep, &file_lock, &glock);
  if (error)
  {
    code = 9;
    goto out_fd;
  }
  error = gpfs_f_lock(filep, glock.cmd, &file_lock);
  if (error)
  {
    code = 11;
    goto out_fd;
  }
  glock.flock.l_type = file_lock.fl_type;
  if (file_lock.fl_type != F_UNLCK)
  {
    error = posix2glock(&glock, &file_lock);
    if (error)
    {
      code = 13;
      goto out_fd;
    }
  }
  if (copy_to_user(l, &glock, sizeof(glock)))
  {
    code = 15;
    error = -EFAULT;
    goto out_fd;
  }
out_fd:
  FPUT(filep, fput_needed);

out:
  TRACE6(TRACE_VNODE, 3, TRCID_GET_LOCK_EXIT_1,
        "get_lock: exit dfd %d lfd %d cmd %d lock_owner 0x%lX error %d code %d",
         mountdirfd, glock.lfd, glock.cmd, glock.lock_owner, error, code);

  return error;
}

int readlink_by_fh(int mountdirfd, struct gpfs_file_handle __user *fh,
                   char __user *buf, int buffsize)
{
  int error = 0;
  int code = 0;
  struct inode *inode;
  struct dentry *dentry = NULL;

  dentry = dentry_by_handle(mountdirfd, fh);
  if (IS_ERR(dentry))
  {
    code = 2;
    error = PTR_ERR(dentry);
    goto exit;
  }
  inode = dentry->d_inode;
  error = -EINVAL;

  if (inode && inode->i_op && inode->i_op->readlink)
  {
    code = 4;
    error = inode->i_op->readlink(dentry, buf, buffsize);
  }
  dput(dentry);

exit:

  TRACE2(TRACE_VNODE, 3, TRCID_GANESHA_READLINK_FH_EXIT,
         "readlink_by_fh: exit error %d code %d", error, code);

  return error;
}

static int get_kernel_handle(struct gpfs_file_handle __user *ufh,
                             struct gpfs_file_handle __user *fh)
{
  int error = 0;
  int code = 0;
  struct gpfs_file_handle fhandle;

  if (copy_from_user(&fhandle, ufh, FH_HDR_SIZE))
  {
    code = 2;
    error = -EFAULT;
    goto out;
  }
  if ((fhandle.handle_size > OPENHANDLE_HANDLE_LEN) || (fhandle.handle_size <= 0))
  {
    code = 4;
    error = -EINVAL;
    goto out;
  }
  /* copy the full handle */
  if (copy_from_user(fh, ufh, FH_HDR_SIZE + fhandle.handle_size))
  {
    code = 8;
    error = -EFAULT;
    goto out;
  }

out:
  TRACE2(TRACE_VNODE, 3, TRCID_GANESHA_GET_KERNEL_HANDLE_EXIT,
          "get_kernel_handle: exit error %d code %d", error, code);

  return error;
}

int do_sys_stat_by_handle(int mountdirfd, struct gpfs_file_handle __user *ufh,
                          struct kstat *stat)
{
  int retval = 0;
  struct dentry *dentry;
  struct vfsmount *mnt = NULL;
  struct gpfs_file_handle fh;

  retval = get_kernel_handle(ufh, &fh);
  if (retval)
    goto out_handle;

  dentry = handle_to_dentry(mountdirfd, &fh, &mnt);
  if (IS_ERR(dentry))
  {
    retval = PTR_ERR(dentry);
    goto out_handle;
  }
  retval = vfs_getattr(mnt, dentry, stat);

  dput(dentry);
  mntput(mnt);

out_handle:
  return retval;
}

int get_inode_update(int mountdirfd, struct gpfs_file_handle __user *handle,
                     void __user *buf, struct glock __user *glock, int *reasonP,
                     int *flagsP)

{
  int retval = 0;
  struct vfsmount *mnt = NULL;

  if (!capable(CAP_DAC_OVERRIDE))
  {
    retval = -EPERM;
    goto out_err;
  }
  mnt = get_vfsmount_from_fd(mountdirfd);
  if (IS_ERR(mnt))
  {
    retval = PTR_ERR(mnt);
    goto out_err;
  }
  retval = gpfs_wait_inode_update(mnt->mnt_sb, handle, buf, glock, reasonP, flagsP);
  mntput(mnt);

out_err:
  return retval;
}

int thread_update(int mountdirfd, int *reasonP)

{
  int retval = 0;
  struct vfsmount *mnt = NULL;

  if (!capable(CAP_DAC_OVERRIDE))
  {
    retval = -EPERM;
    goto out_err;
  }
  mnt = get_vfsmount_from_fd(mountdirfd);
  if (IS_ERR(mnt))
  {
    retval = PTR_ERR(mnt);
    goto out_err;
  }
  retval = gpfs_wait_update(mnt->mnt_sb, reasonP);
  mntput(mnt);

out_err:
  return retval;
}

static int link_by_fh(int mountdirfd, struct gpfs_file_handle __user *dir_fh,
                struct gpfs_file_handle __user *dst_fh,
                const char __user *newname, int len)
{
  char *name;
  int error = 0;
  int code = 0;
  struct dentry *new_dentry = NULL;
  struct dentry *dir_dentry = NULL;
  struct dentry *dst_dentry = NULL;
  struct vfsmount *mnt = NULL;

  name = getname(newname);
  if ((name == NULL) || IS_ERR(name))
  {
    code = 2;
    error = -EFAULT;
    goto exit;
  }
  dst_dentry = dentry_by_handle(mountdirfd, dst_fh);
  if (IS_ERR(dst_dentry))
  {
    code = 4;
    error = PTR_ERR(dst_dentry);
    goto put_name;
  }
  dir_dentry = dentry_by_handle(mountdirfd, dir_fh);
  if (IS_ERR(dir_dentry))
  {
    code = 6;
    error = PTR_ERR(dir_dentry);
    goto put_dst_dentry;
  }
  mutex_lock(&dir_dentry->d_inode->i_mutex);
  new_dentry = lookup_one_len(name, dir_dentry, len);
  mutex_unlock(&dir_dentry->d_inode->i_mutex);
  if (IS_ERR(new_dentry))
  {
    code = 8;
    error = PTR_ERR(new_dentry);
    goto put_dir_dentry;
  }
  mnt = get_vfsmount_from_fd(mountdirfd);
  if (IS_ERR(mnt))
  {
    error = PTR_ERR(mnt);
    goto put_new_dentry;
  }
  error = VFS_LINK(mnt, dst_dentry, dir_dentry->d_inode, new_dentry);
  mntput(mnt);

put_new_dentry:
  dput(new_dentry);

put_dir_dentry:
  dput(dir_dentry);

put_dst_dentry:
  dput(dst_dentry);

put_name:
  putname(name);

exit:
  TRACE2(TRACE_VNODE, 3, TRCID_GANESHA_LINK_FH_EXIT,
         "link_by_fh: exit error %d code %d", error, code);

  return error;
}

static int rename_by_fh(int mountdirfd, struct gpfs_file_handle __user *old_fh,
                       const char __user *old_name, int old_len,
                       struct gpfs_file_handle __user *new_fh,
                       const char __user *new_name, int new_len)
{
  char *newname;
  char *oldname;
  int error = 0;
  int code = 0;
  struct vfsmount *mnt = NULL;
  struct dentry *new_dentry = NULL;
  struct dentry *old_dentry = NULL;
  struct dentry *lock_dentry = NULL;
  struct dentry *newdentry = NULL;
  struct dentry *olddentry = NULL;

  oldname = getname(old_name);
  if ((oldname == NULL) || IS_ERR(oldname))
  {
    code = 2;
    error = -EFAULT;
    goto exit;
  }
  newname = getname(new_name);
  if ((newname == NULL) || IS_ERR(newname))
  {
    code = 4;
    error = -EFAULT;
    goto put_oldname;
  }
  old_dentry = dentry_by_handle(mountdirfd, old_fh);
  if (IS_ERR(old_dentry))
  {
    code = 6;
    error = PTR_ERR(old_dentry);
    goto put_newname;
  }
  new_dentry = dentry_by_handle(mountdirfd, new_fh);
  if (IS_ERR(new_dentry))
  {
    code = 8;
    error = PTR_ERR(new_dentry);
    goto put_old_dentry;
  }
  lock_dentry = lock_rename(new_dentry, old_dentry);

  olddentry = lookup_one_len(old_name, old_dentry, old_len);
  if (IS_ERR(olddentry))
  {
    code = 10;
    error = PTR_ERR(olddentry);
    goto put_new_dentry;
  }
  if (!olddentry->d_inode)
  {
    code = 12;
    error = -ENOENT;
    goto put_olddentry;
  }
  newdentry = lookup_one_len(new_name, new_dentry, new_len);
  if (IS_ERR(newdentry))
  {
    code = 14;
    error = PTR_ERR(newdentry);
    goto put_olddentry;
  }
  if (newdentry == lock_dentry)
  {
    code = 16;
    error = -ENOTEMPTY;
    goto put_olddentry;
  }
  mnt = get_vfsmount_from_fd(mountdirfd);
  if (IS_ERR(mnt))
  {
    error = PTR_ERR(mnt);
    goto put_newdentry;
  }
  error = VFS_RENAME(mnt, old_dentry->d_inode, olddentry,
                     new_dentry->d_inode, newdentry);
  mntput(mnt);

put_newdentry:
  dput(newdentry);

put_olddentry:
  dput(olddentry);

put_new_dentry:
  unlock_rename(new_dentry, old_dentry);
  dput(new_dentry);

put_old_dentry:
  dput(old_dentry);

put_newname:
  putname(newname);

put_oldname:
  putname(oldname);

exit:
  TRACE2(TRACE_VNODE, 3, TRCID_GANESHA_RENAME_FH_EXIT,
         "rename_by_fh: exit error %d code %d", error, code);

  return error;
}

/* If this is a close for Ganesha that we want to keep the locks for fl_owner is NULL */
static int close_file(struct close_file_arg *arg)
{
  int retval = 0;
  fl_owner_t owner;
  struct file *filp;
  struct inode *inode;
  int fput_needed;

  TRACE4(TRACE_VNODE, 3, TRCID_GANESHA_CLOSE_FILE,
        "close_file: dirfd %d fd %d flags %d lock_owner 0x%lX",
         arg->mountdirfd, arg->close_fd, arg->close_flags, arg->close_owner);

  filp = FGET(arg->close_fd, fput_needed);
  if (!filp)
  {
    retval = -EBADF;
    goto out_err;
  }
  if (arg->close_owner != NULL)
    owner = arg->close_owner+1;  // make sure it is not a valid address.
  else
    owner = arg->close_owner;

  retval = filp_close(filp, owner);

  FPUT(filp, fput_needed);

out_err:

  return retval;
}

#if BITS_PER_LONG != 64
static long cp_new_stat64(struct kstat *stat, struct stat64 __user *statbuf)
{
  struct stat64 tmp;

#ifdef CONFIG_MIPS
  /* mips has weird padding, so we don't get 64 bits there */
  if (!new_valid_dev(stat->dev) || !new_valid_dev(stat->rdev))
    return -EOVERFLOW;

  tmp.st_dev = new_encode_dev(stat->dev);
  tmp.st_rdev = new_encode_dev(stat->rdev);
#else
  tmp.st_dev = huge_encode_dev(stat->dev);
  tmp.st_rdev = huge_encode_dev(stat->rdev);
#endif
  tmp.st_ino = stat->ino;
  if (sizeof(tmp.st_ino) < sizeof(stat->ino) && tmp.st_ino != stat->ino)
    return -EOVERFLOW;
#ifdef STAT64_HAS_BROKEN_ST_INO
  tmp.__st_ino = stat->ino;
#endif
  tmp.st_mode = stat->mode;
  tmp.st_nlink = stat->nlink;
  tmp.st_uid = stat->uid;
  tmp.st_gid = stat->gid;
  tmp.st_atime = stat->atime.tv_sec;
  tmp.st_atime_nsec = stat->atime.tv_nsec;
  tmp.st_mtime = stat->mtime.tv_sec;
  tmp.st_mtime_nsec = stat->mtime.tv_nsec;
  tmp.st_ctime = stat->ctime.tv_sec;
  tmp.st_ctime_nsec = stat->ctime.tv_nsec;
  tmp.st_size = stat->size;
  tmp.st_blocks = stat->blocks;
  tmp.st_blksize = stat->blksize;
  return copy_to_user(statbuf, &tmp,sizeof(tmp)) ? -EFAULT : 0;
}

int stat_by_handle(int mountdirfd, struct gpfs_file_handle __user *handle,
                    struct stat64 __user *buf)
{
  long ret;
  struct kstat stat;

  if (!capable(CAP_DAC_OVERRIDE))
    return -EPERM;

  ret = do_sys_stat_by_handle(mountdirfd, handle, &stat);
  if (ret)
    return ret;

  ret = cp_new_stat64(&stat, buf);
  return ret;
}
#else

static int cp_new_stat(struct kstat *stat, struct stat __user *statbuf)
{
  struct stat tmp;

#if BITS_PER_LONG == 32
  if (!old_valid_dev(stat->dev) || !old_valid_dev(stat->rdev))
    return -EOVERFLOW;
#else
  if (!new_valid_dev(stat->dev) || !new_valid_dev(stat->rdev))
    return -EOVERFLOW;
#endif

#if BITS_PER_LONG == 32
  tmp.st_dev = old_encode_dev(stat->dev);
#else
  tmp.st_dev = new_encode_dev(stat->dev);
#endif
  tmp.st_ino = stat->ino;
  if (sizeof(tmp.st_ino) < sizeof(stat->ino) && tmp.st_ino != stat->ino)
    return -EOVERFLOW;

  tmp.st_mode = stat->mode;
  tmp.st_nlink = stat->nlink;
  if (tmp.st_nlink != stat->nlink)
    return -EOVERFLOW;

  SET_UID(tmp.st_uid, stat->uid);
  SET_GID(tmp.st_gid, stat->gid);
#if BITS_PER_LONG == 32
  tmp.st_rdev = old_encode_dev(stat->rdev);
#else
  tmp.st_rdev = new_encode_dev(stat->rdev);
#endif
#if BITS_PER_LONG == 32
  if (stat->size > MAX_NON_LFS)
    return -EOVERFLOW;
#endif
  tmp.st_size = stat->size;
  tmp.st_atime = stat->atime.tv_sec;
  tmp.st_mtime = stat->mtime.tv_sec;
  tmp.st_ctime = stat->ctime.tv_sec;
#ifdef STAT_HAVE_NSEC
  tmp.st_atime_nsec = stat->atime.tv_nsec;
  tmp.st_mtime_nsec = stat->mtime.tv_nsec;
  tmp.st_ctime_nsec = stat->ctime.tv_nsec;
#endif
  tmp.st_blocks = stat->blocks;
  tmp.st_blksize = stat->blksize;
  return copy_to_user(statbuf, &tmp,sizeof(tmp)) ? -EFAULT : 0;
}

#endif
static int stat_by_name(int mountdirfd, struct gpfs_file_handle __user *dir_fh,
                const char __user *newname, int len,
                struct stat __user *buf)
{
  char *name;
  int error = 0;
  int code = 0;
  struct dentry *new_dentry = NULL;
  struct dentry *dir_dentry = NULL;
  struct vfsmount *mnt = NULL;
  struct kstat stat;

  name = getname(newname);
  if ((name == NULL) || IS_ERR(name))
  {
    code = 2;
    error = -EFAULT;
    goto exit;
  }
  dir_dentry = dentry_by_handle(mountdirfd, dir_fh);
  if (IS_ERR(dir_dentry))
  {
    code = 4;
    error = PTR_ERR(dir_dentry);
    goto put_name;
  }
  mutex_lock(&dir_dentry->d_inode->i_mutex);
  new_dentry = lookup_one_len(name, dir_dentry, len);
  mutex_unlock(&dir_dentry->d_inode->i_mutex);
  if (IS_ERR(new_dentry))
  {
    code = 6;
    error = PTR_ERR(new_dentry);
    goto put_dir_dentry;
  }
  if (new_dentry->d_inode == NULL)
  {
    code = 8;
    error = -ENOENT;
    goto put_new_dentry;
  }
  mnt = get_vfsmount_from_fd(mountdirfd);
  if (IS_ERR(mnt))
  {
    code = 10;
    error = PTR_ERR(mnt);
    goto put_new_dentry;
  }
  error = vfs_getattr(mnt, new_dentry, &stat);
  if (error)
    goto put_mnt;

#if BITS_PER_LONG != 64
  error = cp_new_stat64(&stat, buf);
#else
  error = cp_new_stat(&stat, buf);
#endif

put_mnt:
  mntput(mnt);

put_new_dentry:
  dput(new_dentry);

put_dir_dentry:
  dput(dir_dentry);

put_name:
  putname(name);

exit:
  TRACE2(TRACE_VNODE, 3, TRCID_GANESHA_STAT_BY_NAME_EXIT,
         "stat_by_name: exit error %d code %d", error, code);

  return error;
}

static int get_handle(int mountdirfd, struct gpfs_file_handle __user *dir_fh,
                     struct gpfs_file_handle __user *out_fh,
                     const char __user *newname, int len)
{
  char *name;
  int error = 0;
  int code = 0;
  struct dentry *new_dentry = NULL;
  struct dentry *dir_dentry = NULL;

  name = getname(newname);
  if ((name == NULL) || IS_ERR(name))
  {
    code = 2;
    error = -EFAULT;
    goto exit;
  }
  dir_dentry = dentry_by_handle(mountdirfd, dir_fh);
  if (IS_ERR(dir_dentry))
  {
    code = 4;
    error = PTR_ERR(dir_dentry);
    goto put_name;
  }
  mutex_lock(&dir_dentry->d_inode->i_mutex);
  new_dentry = lookup_one_len(name, dir_dentry, len);
  mutex_unlock(&dir_dentry->d_inode->i_mutex);
  if (IS_ERR(new_dentry))
  {
    code = 6;
    error = PTR_ERR(new_dentry);
    goto put_dir_dentry;
  }
  if (new_dentry->d_inode == NULL)
  {
    code = 8;
    error = -ENOENT;
    goto put_new_dentry;
  }
  error = dentry_to_handle(new_dentry, out_fh);

put_new_dentry:
  dput(new_dentry);

put_dir_dentry:
  dput(dir_dentry);

put_name:
  putname(name);

exit:
  TRACE2(TRACE_VNODE, 3, TRCID_GANESHA_GET_FH_EXIT,
         "get_handle: exit error %d code %d", error, code);

  return error;
}

long get_acl_by_handle(struct xstat_arg *args)
{
  struct kstat stat;
  int retval = 0;
  int error = 0;
  struct dentry *dentry;
  struct vfsmount *mnt = NULL;
  struct gpfs_file_handle fh;

  if (!capable(CAP_DAC_OVERRIDE))
    return -EPERM;

  retval = get_kernel_handle(args->handle, &fh);
  if (retval)
    return retval;

  dentry = handle_to_dentry(args->mountdirfd, &fh, &mnt);
  if (IS_ERR(dentry))
  {
    retval = PTR_ERR(dentry);
    goto out_handle;
  }
  if (args->acl != NULL && (args->attr_valid & XATTR_ACL))
    error = vfs_getacl(dentry, args->acl);
  if (args->buf != NULL && (args->attr_valid & XATTR_STAT))
  {
    retval = vfs_getattr(mnt, dentry, &stat);
    if (retval)
      goto out_mnt;
#if BITS_PER_LONG != 64
    retval = cp_new_stat64(&stat, args->buf);
#else
    retval = cp_new_stat(&stat, args->buf);
#endif
  }
out_mnt:
  dput(dentry);
  mntput(mnt);
  if (!retval)
    retval = error;
out_handle:
  return retval;
}

long set_acl_by_handle(struct xstat_arg *args)
{
#if BITS_PER_LONG != 64
  struct stat64 tmp;
#else
  struct stat tmp;
#endif
  int retval = 0;
  int code = 0;
  struct dentry *dentry;
  struct vfsmount *mnt = NULL;
  struct gpfs_file_handle fh;
  struct iattr iattr;
  struct iattr *iattrP = &iattr;

  if (!capable(CAP_DAC_OVERRIDE))
    return -EPERM;

  retval = get_kernel_handle(args->handle, &fh);
  if (retval)
    return retval;

  dentry = handle_to_dentry(args->mountdirfd, &fh, &mnt);
  if (IS_ERR(dentry))
  {
    retval = PTR_ERR(dentry);
    code = 2;
    goto out_handle;
  }
  if (args->acl != NULL && (args->attr_valid & XATTR_ACL))
  {
    retval = vfs_setacl(dentry, args->acl);
    if (retval)
    {
      code = 4;
      goto out_mnt;
    }
  }
  if (args->buf != NULL && (args->attr_valid & XATTR_STAT))
  {
#if BITS_PER_LONG != 64
    if (copy_from_user((void *)&tmp, args->buf, sizeof(struct stat64)))
#else
    if (copy_from_user((void *)&tmp, args->buf, sizeof(struct stat)))
#endif
    {
      code = 6;
      retval = -EFAULT;
      goto out_mnt;
    }
    iattrP->ia_valid = 0;
    if (args->attr_changed & XATTR_MODE)
    {
      iattrP->ia_valid |= ATTR_MODE;
      iattrP->ia_mode = tmp.st_mode;
    }
    if (args->attr_changed & XATTR_UID)
    {
      iattrP->ia_valid |= ATTR_UID;
      iattrP->ia_uid = tmp.st_uid;
    }
    if (args->attr_changed & XATTR_GID)
    {
      iattrP->ia_valid |= ATTR_GID;
      iattrP->ia_gid = tmp.st_gid;
    }
    if (args->attr_changed & XATTR_SIZE)
    {
      iattrP->ia_valid |= ATTR_SIZE;
      iattrP->ia_size = tmp.st_size;
    }
    if (args->attr_changed & XATTR_ATIME)
    {
      iattrP->ia_valid |= ATTR_ATIME;
      iattrP->ia_atime.tv_sec = tmp.st_atime;
#ifdef STAT_HAVE_NSEC
      iattrP->ia_atime.tv_nsec = tmp.st_atime_nsec;
#else
      iattrP->ia_atime.tv_nsec = 0;
#endif
    }
    if (args->attr_changed & XATTR_CTIME)
    {
      iattrP->ia_valid |= ATTR_CTIME;
      iattrP->ia_ctime.tv_sec = tmp.st_ctime;
#ifdef STAT_HAVE_NSEC
      iattrP->ia_ctime.tv_nsec = tmp.st_ctime_nsec;
#else
      iattrP->ia_ctime.tv_nsec = 0;
#endif
    }
    if (args->attr_changed & XATTR_MTIME)
    {
      iattrP->ia_valid |= ATTR_MTIME;
      iattrP->ia_mtime.tv_sec = tmp.st_mtime;
#ifdef STAT_HAVE_NSEC
      iattrP->ia_mtime.tv_nsec = tmp.st_mtime_nsec;
#else
      iattrP->ia_mtime.tv_nsec = 0;
#endif
    }
    if (args->attr_changed & XATTR_ATIME_SET)
      iattrP->ia_valid |= ATTR_ATIME_SET;
    if (args->attr_changed & XATTR_MTIME_SET)
      iattrP->ia_valid |= ATTR_MTIME_SET;

    retval = gpfs_i_setattr(dentry, iattrP);
    if (retval)
    {
      code = 10;
      goto out_mnt;
    }
  }

out_mnt:
  dput(dentry);
  mntput(mnt);

out_handle:

  TRACE3(TRACE_VNODE, 3, TRCID_GANESHA_SET_ACL_EXIT,
         "set_acl_by_handle: exit mountdirfd %d retval %d code %d",
          args->mountdirfd, retval, code);

  return retval;
}

long access_by_handle(int mountdirfd, struct gpfs_file_handle __user *uhandle,
                      struct xstat_cred_t __user *cred, int posix_mode,
                      int access, int *supp)
{
  int i, retval = 0;
  struct dentry *dentry;
  struct vfsmount *mnt = NULL;
  struct gpfs_file_handle fh;
  struct xstat_cred_t uCred;
  ext_cred_t eCred;
  ext_cred_t *eCredP = &eCred;
  struct gpfsVfsData_t *privVfsP;
  cxiNode_t *cnP;
  struct inode *iP;
  int supported;

  if (!capable(CAP_DAC_OVERRIDE))
    return -EPERM;

  retval = get_kernel_handle(uhandle, &fh);
  if (retval)
    return retval;

  dentry = handle_to_dentry(mountdirfd, &fh, &mnt);
  if (IS_ERR(dentry))
  {
    retval = PTR_ERR(dentry);
    goto out_handle;
  }
  iP = dentry->d_inode;
  cnP = VP_TO_CNP(iP);
  privVfsP = VP_TO_PVP(iP);

  if (access)        /* call permission check only if got access mode */
  {
    if (copy_from_user((void *)&uCred, cred, sizeof(uCred)))
    {
      retval = -EFAULT;
      goto out_mnt;
    }
    retval = getCred(&eCred, &eCredP);
    if (retval)
    {
      retval = -ENOMEM;
      goto out_mnt;
    }
    eCredP->principal = uCred.principal;
    eCredP->group = uCred.group;
    eCredP->capabilities = 0;
    eCredP->num_groups = uCred.num_groups;
    if (eCredP->num_groups > ECRED_NGROUPS)
    {
      putCred(&eCred, &eCredP);
      retval = -EINVAL;
      goto out_mnt;
    }
    for(i = 0; i < ECRED_NGROUPS; i++)
      eCredP->eGroups[i] = uCred.eGroups[i];

    retval = gpfs_ops.gpfsAccess(privVfsP, cnP, posix_mode, ACC_SELF,
                                 &access, NULL, eCredP);
    putCred(&eCred, &eCredP);

    if (retval)
      retval = -EACCES;
  }
  supported = 0;
  if (copy_to_user(supp, &supported, sizeof(int)))
  {
    retval = -EFAULT;
    goto out_mnt;
  }

out_mnt:
  dput(dentry);
  mntput(mnt);

out_handle:
  return retval;
}

int set_share_reserve(struct share_reserve_arg *sarg)
{
  int retval = 0;

  TRACE4(TRACE_VNODE, 3, TRCID_LINUXOPS_SHARE_RESERVE_IN,
        "set_share_reserve: dirfd %d fd %d access 0x%X deny 0x%X",
         sarg->mountdirfd, sarg->openfd, sarg->share_access, sarg->share_deny);

  retval = kxGetShare(sarg->openfd, sarg->share_access, sarg->share_deny);

  TRACE3(TRACE_VNODE, 3, TRCID_LINUXOPS_SHARE_RESERVE_EXIT,
         "set_share_reserve exit: retval %d -retval %d fd %d",
          retval, -(retval), sarg->openfd);

  return retval;
}

static int handle_layout_commit(struct layoutcommit_arg *targ)
{
  int retval = 0;
  cxiNode_t *cnP;
  ext_cred_t eCred;
  ext_cred_t *eCredP = &eCred;
  struct gpfsVfsData_t *privVfsP;
  struct MMFSVInfo *vinfoP;
  struct vfsmount *mnt;
  struct dentry *dentry;
  struct gpfs_file_handle fh;
  struct inode *iP = NULL;

  if (!capable(CAP_DAC_OVERRIDE))
    return -EPERM;

  retval = get_kernel_handle(targ->handle, &fh);
  if (retval)
    goto out_err;

  dentry = handle_to_dentry(targ->mountdirfd, &fh, &mnt);
  if (IS_ERR(dentry))
  {
    retval = PTR_ERR(dentry);
    goto out_err;
  }
  iP = dentry->d_inode;
  DBGASSERT(iP != NULL);

  cnP = VP_TO_CNP(iP);
  privVfsP = VP_TO_PVP(iP);
  DBGASSERT(privVfsP != NULL);

  retval = getCred(&eCred, &eCredP);
  if (retval)
    goto out_mnt;

  retval = gpfs_ops.gpfsFsync(privVfsP, NULL, cnP, FFILESYNC, eCredP);
  putCred(&eCred, &eCredP);

out_mnt:
  dput(dentry);
  mntput(mnt);

out_err:

  TRACE1(TRACE_VNODE, 3, TRCID_GANESHA_HANDLE_LAYOUT_COMMIT_EXIT,
         "handle_layout_commit: exit retval %d",
          retval);

  return retval;
}

static int handle_fsync(struct fsync_arg *farg)
{
  int retval = 0;
  cxiNode_t *cnP;
  ext_cred_t eCred;
  ext_cred_t *eCredP = &eCred;
  struct gpfsVfsData_t *privVfsP;
  struct MMFSVInfo *vinfoP;
  struct vfsmount *mnt;
  struct dentry *dentry;
  struct gpfs_file_handle fh;
  struct inode *iP = NULL;
  int flags = FFILESYNC;
  uint32_t verifier[2];
  uint32_t *verifierP = (uint32_t *)&verifier;

  if (!capable(CAP_DAC_OVERRIDE))
    return -EPERM;

  retval = get_kernel_handle(farg->handle, &fh);
  if (retval)
    goto out_err;

  dentry = handle_to_dentry(farg->mountdirfd, &fh, &mnt);
  if (IS_ERR(dentry))
  {
    retval = PTR_ERR(dentry);
    goto out_err;
  }
  iP = dentry->d_inode;
  DBGASSERT(iP != NULL);

  cnP = VP_TO_CNP(iP);
  privVfsP = VP_TO_PVP(iP);
  DBGASSERT(privVfsP != NULL);

  retval = getCred(&eCred, &eCredP);
  if (retval)
    goto out_mnt;

  retval = gpfs_ops.gpfsFsyncRange(privVfsP, cnP, flags, farg->offset,
                                   farg->length, eCredP);
  putCred(&eCred, &eCredP);
  if (retval)
    goto out_mnt;

#ifdef P_NFS4
  gpfs_ops.gpfsGetVerifier(privVfsP, verifierP);
  if (copy_to_user(farg->verifier4, verifierP, sizeof(verifier)))
    retval = -EFAULT;
#endif

out_mnt:
  dput(dentry);
  mntput(mnt);

out_err:

  TRACE1(TRACE_VNODE, 3, TRCID_GANESHA_HANDLE_FSYNC_EXIT,
         "handle_fsync: exit retval %d", retval);

  return retval;
}


static int handle_layout_type(int mountdirfd, struct gpfs_file_handle *handle)
{
  int retval = 0;
  struct vfsmount *mnt;

  mnt = get_vfsmount_from_fd(mountdirfd);
  if (IS_ERR(mnt))
  {
    retval = PTR_ERR(mnt);
    goto out_err;
  }
#ifdef P_NFS4
  retval = gpfs_layout_type(mnt->mnt_sb);
#else
  retval = 0;
#endif
  mntput(mnt);

out_err:

  TRACE2(TRACE_VNODE, 3, TRCID_GANESHA_HANDLE_LAYOUT_TYPE_EXIT,
         "handle_layout_type: exit mountdirfd %d retval %d",
          mountdirfd, retval);

  return retval;
}

static int handle_get_layout(struct layoutget_arg *yarg)
{
  int retval = 0;
  struct gpfs_file_handle __user *uhandle = yarg->handle;
  struct vfsmount *mnt;
  struct dentry *dentry;
  struct gpfs_file_handle fh;
  struct inode *iP = NULL;
  struct nfsd4_pnfs_layoutget_res res;
  struct pnfs_filelayout_layout file_layout;
  int mdsAddr = 0;
  unsigned int *fhP;

  if (!capable(CAP_DAC_OVERRIDE))
    return -EPERM;

  retval = get_kernel_handle(uhandle, &fh);
  if (retval)
    goto out_err;

  dentry = handle_to_dentry(yarg->mountdirfd, &fh, &mnt);
  if (IS_ERR(dentry))
  {
    retval = PTR_ERR(dentry);
    goto out_err;
  }
  iP = dentry->d_inode;
#ifdef P_NFS4
  fhP = (int *)&(fh.f_handle);
  mdsAddr = fhP[6];
  printk("handle_get_layout 1 last 4 bytes 0x%X\n", mdsAddr);

  retval = gpfs_layout_get(iP, (void *)yarg->xdr, &yarg->args,
                               (struct nfsd4_pnfs_layoutget_res *)&file_layout);
  if (retval == 0)
  {
    mdsAddr = file_layout.device_id.sbid;
    fhP[6] = mdsAddr;
    /* copy the full handle back */
    if (copy_to_user(uhandle, &fh, FH_HDR_SIZE + fh.handle_size))
    {
      retval = -EFAULT;
      goto out_mnt;
    }
    if (copy_to_user(yarg->file_layout, &file_layout, sizeof(file_layout)))
    {
      retval = -EFAULT;
      goto out_mnt;
    }
  }
#else
  retval = -ENOSYS;
#endif

out_mnt:
  dput(dentry);
  mntput(mnt);

out_err:

  TRACE3(TRACE_VNODE, 3, TRCID_GANESHA_HANDLE_GET_LAYOUT_EXIT,
         "handle_get_layout: exit mountdirfd %d retval %d mdsAddr 0x%X",
          yarg->mountdirfd, retval, mdsAddr);

  return retval;
}

static int handle_return_layout(struct layoutreturn_arg *zarg)
{
  int retval = 0;
  struct gpfs_file_handle __user *uhandle = zarg->handle;
  struct vfsmount *mnt;
  struct dentry *dentry;
  struct gpfs_file_handle fh;
  struct nfsd4_pnfs_layoutreturn_arg lr;
  struct inode *iP = NULL;

  if (!capable(CAP_DAC_OVERRIDE))
    return -EPERM;

  retval = get_kernel_handle(uhandle, &fh);
  if (retval)
    goto out_err;

  dentry = handle_to_dentry(zarg->mountdirfd, &fh, &mnt);
  if (IS_ERR(dentry))
  {
    retval = PTR_ERR(dentry);
    goto out_err;
  }
  iP = dentry->d_inode;
#ifdef P_NFS4
  retval = gpfs_layout_return(iP, &zarg->args);
#else
  retval = -ENOSYS;
#endif

out_mnt:
  dput(dentry);
  mntput(mnt);

out_err:

  TRACE2(TRACE_VNODE, 3, TRCID_GANESHA_HANDLE_RETURN_LAYOUT_EXIT,
         "handle_return_layout: exit mountdirfd %d retval %d",
          zarg->mountdirfd, retval);

  return retval;
}

static int handle_dsread(struct dsread_arg *rarg)
{
  cxiNode_t *cnP;
  ext_cred_t eCred;
  ext_cred_t *eCredP = &eCred;
  struct gpfsVfsData_t *privVfsP;
  struct MMFSVInfo *vinfoP;
  int fd, retval = 0;
  int open_flag;
  struct gpfs_file_handle __user *uhandle = rarg->handle;
  struct vfsmount *mnt;
  struct dentry *dentry;
  struct gpfs_file_handle fh;
  struct inode *iP = NULL;
  struct file *filp;
  uint64_t offset = rarg->offset;
  int mdsAddr = 0;
  unsigned int *fhP;

  if (!capable(CAP_DAC_OVERRIDE))
    return -EPERM;

  retval = get_kernel_handle(uhandle, &fh);
  if (retval)
    goto out;

  fhP = (int *)&(fh.f_handle);
  mdsAddr = fhP[6];

  dentry = handle_to_dentry(rarg->mountdirfd, &fh, &mnt);
  if (IS_ERR(dentry))
  {
    retval = PTR_ERR(dentry);
    goto out;
  }
  fd = get_unused_fd();
  if (fd < 0)
  {
    retval = fd;
    goto out_dentry;
  }
#if 0 //??? direct IO conflicts with exclusive-share token
  open_flag = FREAD | O_DIRECT;
#else
  open_flag = FREAD;
#endif
  filp = DENTRY_OPEN(dentry, mnt, open_flag);
  if (IS_ERR(filp))
  {
    put_unused_fd(fd);
    retval = PTR_ERR(filp);
    goto out_dentry;
  }
  iP = dentry->d_inode;
  cnP = VP_TO_CNP(iP);
  privVfsP = VP_TO_PVP(iP);
  DBGASSERT(privVfsP != NULL);
  vinfoP = (struct MMFSVInfo *)filp->private_data;

  retval = getCred(&eCred, &eCredP);
  if (retval)
    goto out;

#ifdef P_NFS4
  retval = gpfs_ops.gpfsGetOpenState(privVfsP, vinfoP, cnP, mdsAddr, NULL, 0, eCredP);
  putCred(&eCred, &eCredP);
  if (retval)
    goto out;
#endif

  retval = gpfs_f_read(filp, rarg->bufP, rarg->length, &offset);

  fput(filp);

out:

  TRACE3(TRACE_VNODE, 3, TRCID_GANESHA_HANDLE_DS_READ_EXIT,
         "handle_dsread: exit mountdirfd %d retval %d mdsAddr 0x%X",
          rarg->mountdirfd, retval, mdsAddr);

  return retval;

out_dentry:
  dput(dentry);
  mntput(mnt);
  goto out;
}

static int handle_dswrite(struct dswrite_arg *warg)
{
  cxiNode_t *cnP;
  ext_cred_t eCred;
  ext_cred_t *eCredP = &eCred;
  int fd, retval = 0;
  int open_flag;
  struct gpfs_file_handle __user *uhandle = warg->handle;
  struct vfsmount *mnt;
  struct dentry *dentry;
  struct gpfs_file_handle fh;
  struct inode *iP = NULL;
  struct file *filp;
  struct gpfsVfsData_t *privVfsP;
  struct MMFSVInfo *vinfoP;
  uint64_t offset = warg->offset;
  int mdsAddr = 0;
  uint32_t stability = x_UNSTABLE4;
  uint32_t verifier[2];
  uint32_t *verifierP = (uint32_t *)&verifier;
  unsigned int *fhP;

  if (!capable(CAP_DAC_OVERRIDE))
    return -EPERM;

  retval = get_kernel_handle(uhandle, &fh);
  if (retval)
    goto out;

  fhP = (int *)&(fh.f_handle);
  mdsAddr = fhP[6];

  dentry = handle_to_dentry(warg->mountdirfd, &fh, &mnt);
  if (IS_ERR(dentry))
  {
    retval = PTR_ERR(dentry);
    goto out;
  }
  fd = get_unused_fd();
  if (fd < 0)
  {
    retval = fd;
    goto out_dentry;
  }
#if 0 //??? direct IO conflicts with exclusive-share token
  open_flag = FWRITE | O_DIRECT;
#else
  open_flag = FWRITE;
#endif
  filp = DENTRY_OPEN(dentry, mnt, open_flag);
  if (IS_ERR(filp))
  {
    put_unused_fd(fd);
    retval = PTR_ERR(filp);
    goto out_dentry;
  }
  iP = dentry->d_inode;
  cnP = VP_TO_CNP(iP);
  privVfsP = VP_TO_PVP(iP);
  DBGASSERT(privVfsP != NULL);
  vinfoP = (struct MMFSVInfo *)filp->private_data;

  retval = getCred(&eCred, &eCredP);
  if (retval)
    goto out;

#ifdef P_NFS4
  retval = gpfs_ops.gpfsGetOpenState(privVfsP, vinfoP, cnP, mdsAddr, NULL, 0, eCredP);
  putCred(&eCred, &eCredP);
  if (retval)
    goto out;
#endif

  retval = gpfs_f_write(filp, warg->bufP, warg->length, &offset);
  if (retval > 0)
  {
   if (copy_to_user(warg->stability_got, &stability, sizeof(int)))
   {
     retval = -EFAULT;
     goto out_fput;
   }
   privVfsP = VP_TO_PVP(dentry->d_inode);
#ifdef P_NFS4
   gpfs_ops.gpfsGetVerifier(privVfsP, verifierP);
   if (copy_to_user(warg->verifier4, verifierP, sizeof(verifier)))
   {
     retval = -EFAULT;
     goto out_fput;
   }
#endif
  }

out_fput:
  fput(filp);

out:

  TRACE6(TRACE_VNODE, 3, TRCID_GANESHA_HANDLE_DS_WRITE_EXIT,
      "handle_dswrite: exit mountdirfd %d retval %d stability %d mdsAddr 0x%X verifier %d-%d",
       warg->mountdirfd, retval, stability, mdsAddr, verifier[0], verifier[1]);

  return retval;

out_dentry:
  dput(dentry);
  mntput(mnt);
  goto out;
}

static int handle_get_verifier(void *arg)
{
  uint32_t verifier[2];
  uint32_t *verifierP = (uint32_t *)&verifier;

#ifdef P_NFS4
  gpfs_ops.gpfsGetVerifier(NULL, verifierP);
#endif
  if (copy_to_user(arg, verifierP, sizeof(verifier)))
    return -EFAULT;

  return 0;
}

static int handle_get_nodeid(void *arg)
{
  return gpfs_ops.gpfsGetNodeID(NULL);
}

static int handle_get_deviceinfo(struct deviceinfo_arg *darg)
{
  int retval = 0;
  struct vfsmount *mnt;

  mnt = get_vfsmount_from_fd(darg->mountdirfd);
  if (IS_ERR(mnt))
  {
    retval = PTR_ERR(mnt);
    goto out_err;
  }
#ifdef P_NFS4
  retval = gpfs_get_deviceinfo(mnt->mnt_sb, &darg->xdr, darg->type,
                               &darg->devid);
#else
  retval = -ENOSYS;
#endif
  mntput(mnt);

out_err:

  TRACE4(TRACE_VNODE, 3, TRCID_GANESHA_HANDLE_GET_DEVICEINFO_EXIT,
      "handle_get_deviceinfo: exit mountdirfd %d retval 0x%llx p 0x%llx end %p",
       darg->mountdirfd, retval, darg->xdr.p, darg->xdr.end);

  return retval;
}

int kxGanesha(int op, void *arg)
{
  int rc = 0;
  int code = 0;
  int retval = 0;
  all_args args;

  VFS_STAT_START(op);
  ENTER(0);

  TRACE3(TRACE_VNODE, 1, TRCID_LINUX_GANESHA_ENTER,
         "kxGanesha: enter op %d tgid %d gid %d",
          op, cxiGetProcessId(), current->pid);

  /* Only allow the root users */
  if (!cxiIsSuperUser())
  {
    retval = -EPERM;
    goto out;
  }

  switch (op)
  {
    case OPENHANDLE_GET_HANDLE:
      if (copy_from_user(&args.gharg, (void *)arg, sizeof(struct get_handle_arg)))
        return -EFAULT;
      code = 1;
      retval = get_handle(args.gharg.mountdirfd, args.gharg.dir_fh,
                          args.gharg.out_fh, args.gharg.name, args.gharg.len);
      break;

    case OPENHANDLE_NAME_TO_HANDLE:
      if (copy_from_user(&args.harg, (void *)arg, sizeof(struct name_handle_arg)))
        return -EFAULT;
      code = 1;
      retval = name_to_handle_at(args.harg.dfd, args.harg.name,
                                 args.harg.handle, args.harg.flag);
      break;

    case OPENHANDLE_SET_LOCK:
      if (copy_from_user(&args.larg, (void *)arg, sizeof(struct set_get_lock_arg)))
        return -EFAULT;
      retval = set_lock(args.larg.mountdirfd, args.larg.lock, false);
      break;

    case OPENHANDLE_GET_LOCK:
      if (copy_from_user(&args.larg, (void *)arg, sizeof(struct set_get_lock_arg)))
        return -EFAULT;
      retval = get_lock(args.larg.mountdirfd, args.larg.lock);
      break;

    case OPENHANDLE_OPEN_BY_HANDLE:
      if (copy_from_user(&args.oarg, (void *)arg, sizeof(struct open_arg)))
        return -EFAULT;
      retval = open_by_handle(args.oarg.mountdirfd, args.oarg.handle,
                              args.oarg.flags, 0, 0);
      break;

    case OPENHANDLE_OPEN_SHARE_BY_HANDLE:
      if (copy_from_user(&args.osarg, (void *)arg, sizeof(struct open_share_arg)))
        return -EFAULT;
      retval = open_by_handle(args.osarg.mountdirfd, args.osarg.handle,
                              args.osarg.flags, args.osarg.share_access,
                              args.osarg.share_deny);
      break;

    case OPENHANDLE_SHARE_RESERVE:
      if (copy_from_user(&args.sarg, (void *)arg, sizeof(struct share_reserve_arg)))
        return -EFAULT;
      retval = set_share_reserve(&args.sarg);
      break;

    case OPENHANDLE_SET_DELEGATION:
      if (copy_from_user(&args.larg, (void *)arg, sizeof(struct set_get_lock_arg)))
        return -EFAULT;
      retval = set_lock(args.larg.mountdirfd, args.larg.lock, true);
      break;

    case OPENHANDLE_LINK_BY_FH:
      if (copy_from_user(&args.lnarg, (void *)arg, sizeof(struct link_fh_arg)))
        return -EFAULT;
      retval = link_by_fh(args.lnarg.mountdirfd, args.lnarg.dir_fh,
                          args.lnarg.dst_fh, args.lnarg.name, args.lnarg.len);
      break;

    case OPENHANDLE_STAT_BY_NAME:
      if (copy_from_user(&args.snarg, (void *)arg, sizeof(struct stat_name_arg)))
        return -EFAULT;
      retval = stat_by_name(args.snarg.mountdirfd, args.snarg.handle,
                            args.snarg.name, args.snarg.len, args.snarg.buf);
      break;

    case OPENHANDLE_RENAME_BY_FH:
      if (copy_from_user(&args.rnarg, (void *)arg, sizeof(struct rename_fh_arg)))
        return -EFAULT;
      retval = rename_by_fh(args.rnarg.mountdirfd, args.rnarg.old_fh,
                            args.rnarg.old_name, args.rnarg.old_len,
                            args.rnarg.new_fh, args.rnarg.new_name,
                            args.rnarg.new_len);
      break;

    case OPENHANDLE_READLINK_BY_FH:
      if (copy_from_user(&args.rlfharg, (void *)arg,
                                                sizeof(struct readlink_fh_arg)))
        return -EFAULT;
      retval = readlink_by_fh(args.rlfharg.mountdirfd, args.rlfharg.handle,
                              args.rlfharg.buffer, args.rlfharg.size);
      break;

    case OPENHANDLE_INODE_UPDATE:
        if (copy_from_user(&args.barg, (void *)arg, sizeof(struct callback_arg)))
            return -EFAULT;
        retval = get_inode_update(args.barg.mountdirfd, args.barg.handle,
                        (void *)args.barg.buf, args.barg.fl, args.barg.reason,
                        args.barg.flags);
      break;

    case OPENHANDLE_THREAD_UPDATE:
        if (copy_from_user(&args.barg, (void *)arg, sizeof(struct callback_arg)))
            return -EFAULT;
        retval = thread_update(args.barg.mountdirfd, args.barg.reason);
      break;

    case OPENHANDLE_LAYOUT_TYPE:
      if (copy_from_user(&args.oarg, (void *)arg, sizeof(struct open_arg)))
        return -EFAULT;
      retval = handle_layout_type(args.oarg.mountdirfd, args.oarg.handle);
      break;

    case OPENHANDLE_GET_DEVICEINFO:
      if (copy_from_user(&args.darg, (void *)arg, sizeof(struct deviceinfo_arg)))
        return -EFAULT;
      retval = handle_get_deviceinfo(&args.darg);
      break;

    case OPENHANDLE_DS_READ:
      if (copy_from_user(&args.rarg, (void *)arg, sizeof(struct dsread_arg)))
        return -EFAULT;
      retval = handle_dsread(&args.rarg);
      break;

    case OPENHANDLE_DS_WRITE:
      if (copy_from_user(&args.warg, (void *)arg, sizeof(struct dswrite_arg)))
        return -EFAULT;
      retval = handle_dswrite(&args.warg);
      break;

    case OPENHANDLE_GET_VERIFIER:
      retval = handle_get_verifier(arg);
      break;

    case OPENHANDLE_GET_NODEID:
      retval = handle_get_nodeid(arg);
      break;

    case OPENHANDLE_LAYOUT_RETURN:
      if (copy_from_user(&args.zarg, (void *)arg, sizeof(struct layoutreturn_arg)))
        return -EFAULT;
      retval = handle_return_layout(&args.zarg);
      break;

    case OPENHANDLE_LAYOUT_GET:
      if (copy_from_user(&args.yarg, (void *)arg, sizeof(struct layoutget_arg)))
        return -EFAULT;
      retval = handle_get_layout(&args.yarg);
      break;

    case OPENHANDLE_LAYOUT_COMMIT:
      if (copy_from_user(&args.targ, (void *)arg, sizeof(struct layoutcommit_arg)))
        return -EFAULT;
      retval = handle_layout_commit(&args.targ);
      break;

    case OPENHANDLE_FSYNC:
      if (copy_from_user(&args.farg, (void *)arg, sizeof(struct fsync_arg)))
        return -EFAULT;
      retval = handle_fsync(&args.farg);
      break;

    case OPENHANDLE_GET_XSTAT:
      if (copy_from_user(&args.carg, (void *)arg, sizeof(struct xstat_arg)))
        return -EFAULT;
      retval = get_acl_by_handle(&args.carg);
      break;

    case OPENHANDLE_SET_XSTAT:
      if (copy_from_user(&args.carg, (void *)arg, sizeof(struct xstat_arg)))
        return -EFAULT;
      retval = set_acl_by_handle(&args.carg);
      break;

    case OPENHANDLE_CLOSE_FILE:
      if (copy_from_user(&args.cfarg, (void *)arg, sizeof(struct close_file_arg)))
        return -EFAULT;
      retval = close_file((struct close_file_arg *)&args.cfarg);
      break;

    case OPENHANDLE_CHECK_ACCESS:
      if (copy_from_user(&args.xarg, (void *)arg, sizeof(struct xstat_access_arg)))
        return -EFAULT;
      retval = access_by_handle(args.xarg.mountdirfd, args.xarg.handle,
                                args.xarg.cred, args.xarg.posix_mode,
                                args.xarg.access, args.xarg.supported);
      break;

    default:
      retval = -EOPNOTSUPP;
      break;
  }

out:

  TRACE4(TRACE_VNODE, 1, TRCID_LINUX_GANESHA_EXIT,
        "kxGanesha: exit op %d retval %d rc %d code %d", op, retval, rc, code);

  VFS_STAT_STOP;
  EXIT(0);
  return retval;
}
#endif // GANESHA

#ifdef LIGHT_WEIGHT_EVENT
int kxlweCreateSession(lwe_sessid_t oldsid, 
                       char *sessinfop, 
		       lwe_sessid_t *newsidp)
{
  int rc = 0;

  rc = gpfs_ops.gpfslweCreateSession(oldsid, sessinfop, newsidp);
out:
  return rc;
}

int kxlweDestroySession(lwe_sessid_t sid)
{
  int rc = 0;
  rc = gpfs_ops.gpfslweDestroySession(sid);
out:
  return rc;

}

int kxlweGetAllSession(u_int nelem, 
                       lwe_sessid_t *sidbufp, 
		       u_int *nelemp)
{
  int rc = 0;
  rc = gpfs_ops.gpfslweGetAllSessions(nelem, sidbufp, nelemp);
out:
  return rc;

}

int kxlweQuerySession(lwe_sessid_t sid, 
                      size_t buflen, 
		      void* bufp, 
		      size_t *rlenp)
{
  int rc = 0;
  rc = gpfs_ops.gpfslweQuerySession(sid, buflen, bufp, rlenp);
out:
  return rc;

}

int kxlweGetEvents(lwe_sessid_t sid, 
                   unsigned int maxmsgs, 
                   unsigned int flags, 
		   size_t buflen, 
		   void* bufp, 
		   size_t *rlenp)
{
  int rc = 0;
  rc = gpfs_ops.gpfslweGetEvents(sid, maxmsgs, flags, buflen, bufp, rlenp);
out:
  return rc;

}

int kxlweRespondEvent(lwe_sessid_t sid, 
                      lwe_token_t token,
                      lwe_resp_t response, 
		      int reterror)
{
  int rc = 0;
  rc = gpfs_ops.gpfslweRespondEvent(sid, token, response, reterror);
out:
  return rc;

}


#endif /* LIGHT_WEIGHT_EVENT */
